# WARNING: this file is auto-generated by 'async_to_sync.py'
# from the original file 'test_cursor_server_async.py'
# DO NOT CHANGE! Change the original file instead.
import pytest
from packaging.version import parse as ver

import psycopg
from psycopg import errors as e
from psycopg import pq, rows

from ._test_cursor import ph

pytestmark = pytest.mark.crdb_skip("server-side cursor")

cursor_classes = [psycopg.ServerCursor]
# Allow to import (not necessarily to run) the module with psycopg 3.1.
if ver(psycopg.__version__) >= ver("3.2.0.dev0"):
    cursor_classes.append(psycopg.RawServerCursor)


@pytest.fixture(params=cursor_classes)
def conn(conn, request, anyio_backend):
    conn.server_cursor_factory = request.param
    return conn


def test_init_row_factory(conn):
    with psycopg.ServerCursor(conn, "foo") as cur:
        assert cur.name == "foo"
        assert cur.connection is conn
        assert cur.row_factory is conn.row_factory

    conn.row_factory = rows.dict_row

    with psycopg.ServerCursor(conn, "bar") as cur:
        assert cur.name == "bar"
        assert cur.row_factory is rows.dict_row

    with psycopg.ServerCursor(conn, "baz", row_factory=rows.namedtuple_row) as cur:
        assert cur.name == "baz"
        assert cur.row_factory is rows.namedtuple_row


def test_init_params(conn):
    with psycopg.ServerCursor(conn, "foo") as cur:
        assert cur.scrollable is None
        assert cur.withhold is False

    with psycopg.ServerCursor(conn, "bar", withhold=True, scrollable=False) as cur:
        assert cur.scrollable is False
        assert cur.withhold is True


@pytest.mark.crdb_skip("cursor invalid name")
def test_funny_name(conn):
    cur = conn.cursor("1-2-3")
    cur.execute("select generate_series(1, 3) as bar")
    assert cur.fetchall() == [(1,), (2,), (3,)]
    assert cur.name == "1-2-3"
    cur.close()


def test_repr(conn):
    cur = conn.cursor("my-name")
    assert "psycopg.%s" % conn.server_cursor_factory.__name__ in str(cur)
    assert "my-name" in repr(cur)
    cur.close()


def test_connection(conn):
    cur = conn.cursor("foo")
    assert cur.connection is conn
    cur.close()


def test_description(conn):
    cur = conn.cursor("foo")
    assert cur.name == "foo"
    cur.execute("select generate_series(1, 10)::int4 as bar")
    assert len(cur.description) == 1
    assert cur.description[0].name == "bar"
    assert cur.description[0].type_code == cur.adapters.types["int4"].oid
    assert cur.pgresult.ntuples == 0
    cur.close()


def test_format(conn):
    cur = conn.cursor("foo")
    assert cur.format == pq.Format.TEXT
    cur.close()

    cur = conn.cursor("foo", binary=True)
    assert cur.format == pq.Format.BINARY
    cur.close()


def test_query_params(conn):
    with conn.cursor("foo") as cur:
        assert cur._query is None
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (3,))
        assert cur._query is not None
        assert b"declare" in cur._query.query.lower()
        assert b"(1, $1)" in cur._query.query.lower()
        assert cur._query.params == [bytes([0, 3])]  # 3 as binary int2


def test_binary_cursor_execute(conn):
    cur = conn.cursor("foo", binary=True)
    cur.execute("select generate_series(1, 2)::int4")
    assert cur.fetchone() == (1,)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x00\x00\x01"
    assert cur.fetchone() == (2,)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x00\x00\x02"
    cur.close()


def test_execute_binary(conn):
    cur = conn.cursor("foo")
    cur.execute("select generate_series(1, 2)::int4", binary=True)
    assert cur.fetchone() == (1,)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x00\x00\x01"
    assert cur.fetchone() == (2,)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x00\x00\x02"

    cur.execute("select generate_series(1, 1)::int4")
    assert cur.fetchone() == (1,)
    assert cur.pgresult.fformat(0) == 0
    assert cur.pgresult.get_value(0, 0) == b"1"
    cur.close()


def test_binary_cursor_text_override(conn):
    cur = conn.cursor("foo", binary=True)
    cur.execute("select generate_series(1, 2)", binary=False)
    assert cur.fetchone() == (1,)
    assert cur.pgresult.fformat(0) == 0
    assert cur.pgresult.get_value(0, 0) == b"1"
    assert cur.fetchone() == (2,)
    assert cur.pgresult.fformat(0) == 0
    assert cur.pgresult.get_value(0, 0) == b"2"

    cur.execute("select generate_series(1, 2)::int4")
    assert cur.fetchone() == (1,)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x00\x00\x01"
    cur.close()


def test_close(conn, recwarn):
    if conn.info.transaction_status == pq.TransactionStatus.INTRANS:
        # connection dirty from previous failure
        conn.execute("close foo")
    recwarn.clear()
    cur = conn.cursor("foo")
    cur.execute("select generate_series(1, 10) as bar")
    cur.close()
    assert cur.closed

    assert not conn.execute("select * from pg_cursors where name = 'foo'").fetchone()
    del cur
    assert not recwarn, [str(w.message) for w in recwarn.list]


def test_close_idempotent(conn):
    cur = conn.cursor("foo")
    cur.execute("select 1")
    cur.fetchall()
    cur.close()
    cur.close()


def test_close_broken_conn(conn):
    cur = conn.cursor("foo")
    conn.close()
    cur.close()
    assert cur.closed


def test_cursor_close_fetchone(conn):
    cur = conn.cursor("foo")
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    cur.execute(query)
    for _ in range(5):
        cur.fetchone()

    cur.close()
    assert cur.closed

    with pytest.raises(e.InterfaceError):
        cur.fetchone()


def test_cursor_close_fetchmany(conn):
    cur = conn.cursor("foo")
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    cur.execute(query)
    assert len(cur.fetchmany(2)) == 2

    cur.close()
    assert cur.closed

    with pytest.raises(e.InterfaceError):
        cur.fetchmany(2)


def test_cursor_close_fetchall(conn):
    cur = conn.cursor("foo")
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    cur.execute(query)
    assert len(cur.fetchall()) == 10

    cur.close()
    assert cur.closed

    with pytest.raises(e.InterfaceError):
        cur.fetchall()


def test_close_noop(conn, recwarn):
    recwarn.clear()
    cur = conn.cursor("foo")
    cur.close()
    assert not recwarn, [str(w.message) for w in recwarn.list]


def test_close_on_error(conn):
    cur = conn.cursor("foo")
    cur.execute("select 1")
    with pytest.raises(e.ProgrammingError):
        conn.execute("wat")
    assert conn.info.transaction_status == pq.TransactionStatus.INERROR
    cur.close()


def test_pgresult(conn):
    cur = conn.cursor()
    cur.execute("select 1")
    assert cur.pgresult
    cur.close()
    assert not cur.pgresult


def test_context(conn, recwarn):
    recwarn.clear()
    with conn.cursor("foo") as cur:
        cur.execute("select generate_series(1, 10) as bar")

    assert cur.closed
    assert not conn.execute("select * from pg_cursors where name = 'foo'").fetchone()
    del cur
    assert not recwarn, [str(w.message) for w in recwarn.list]


def test_close_no_clobber(conn):
    with pytest.raises(e.DivisionByZero):
        with conn.cursor("foo") as cur:
            cur.execute(ph(cur, "select 1 / %s"), (0,))
            cur.fetchall()


def test_warn_close(conn, recwarn, gc_collect):
    recwarn.clear()
    cur = conn.cursor("foo")
    cur.execute("select generate_series(1, 10) as bar")
    del cur
    gc_collect()
    msg = str(recwarn.pop(ResourceWarning).message)
    assert conn.server_cursor_factory.__name__ in msg
    assert ".close()" in msg


def test_execute_reuse(conn):
    with conn.cursor("foo") as cur:
        query = ph(cur, "select generate_series(1, %s) as foo")
        cur.execute(query, (3,))
        assert cur.fetchone() == (1,)

        query = ph(cur, "select %s::text as bar, %s::text as baz")
        cur.execute(query, ("hello", "world"))
        assert cur.fetchone() == ("hello", "world")
        assert cur.description[0].name == "bar"
        assert cur.description[0].type_code == cur.adapters.types["text"].oid
        assert cur.description[1].name == "baz"


@pytest.mark.parametrize(
    "stmt", ["", "wat", "create table ssc ()", "select 1; select 2"]
)
def test_execute_error(conn, stmt):
    cur = conn.cursor("foo")
    with pytest.raises(e.ProgrammingError):
        cur.execute(stmt)
    cur.close()


def test_executemany(conn):
    cur = conn.cursor("foo")
    with pytest.raises(e.NotSupportedError):
        cur.executemany(ph(cur, "select %s"), [(1,), (2,)])
    cur.close()


def test_fetchone(conn):
    with conn.cursor("foo") as cur:
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (2,))
        assert cur.fetchone() == (1,)
        assert cur.fetchone() == (2,)
        assert cur.fetchone() is None


def test_fetchmany(conn):
    with conn.cursor("foo") as cur:
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (5,))
        assert cur.fetchmany(3) == [(1,), (2,), (3,)]
        assert cur.fetchone() == (4,)
        assert cur.fetchmany(3) == [(5,)]
        assert cur.fetchmany(3) == []


def test_fetchall(conn):
    with conn.cursor("foo") as cur:
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (3,))
        assert cur.fetchall() == [(1,), (2,), (3,)]
        assert cur.fetchall() == []

    with conn.cursor("foo") as cur:
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (3,))
        assert cur.fetchone() == (1,)
        assert cur.fetchall() == [(2,), (3,)]
        assert cur.fetchall() == []


def test_nextset(conn):
    with conn.cursor("foo") as cur:
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (3,))
        assert not cur.nextset()


def test_no_result(conn):
    with conn.cursor("foo") as cur:
        query = ph(cur, "select generate_series(1, %s) as bar where false")
        cur.execute(query, (3,))
        assert len(cur.description) == 1
        assert cur.fetchall() == []


@pytest.mark.parametrize("row_factory", ["tuple_row", "dict_row", "namedtuple_row"])
def test_standard_row_factory(conn, row_factory):
    if row_factory == "tuple_row":
        getter = lambda r: r[0]  # noqa: E731
    elif row_factory == "dict_row":
        getter = lambda r: r["bar"]  # noqa: E731
    elif row_factory == "namedtuple_row":
        getter = lambda r: r.bar  # noqa: E731
    else:
        assert False, row_factory

    row_factory = getattr(rows, row_factory)
    with conn.cursor("foo", row_factory=row_factory) as cur:
        cur.execute("select generate_series(1, 5) as bar")
        assert getter(cur.fetchone()) == 1
        assert list(map(getter, cur.fetchmany(2))) == [2, 3]
        assert list(map(getter, cur.fetchall())) == [4, 5]


@pytest.mark.crdb_skip("scroll cursor")
def test_row_factory(conn):
    n = 0

    def my_row_factory(cur):
        nonlocal n
        n += 1
        return lambda values: [n] + [-v for v in values]

    cur = conn.cursor("foo", row_factory=my_row_factory, scrollable=True)
    cur.execute("select generate_series(1, 3) as x")
    assert cur.rownumber == 0
    recs = cur.fetchall()
    assert cur.rownumber == 3
    cur.scroll(0, "absolute")
    assert cur.rownumber == 0
    while rec := cur.fetchone():
        recs.append(rec)
    assert cur.rownumber == 3
    assert recs == [[1, -1], [1, -2], [1, -3]] * 2

    cur.scroll(0, "absolute")
    cur.row_factory = rows.dict_row
    assert cur.fetchone() == {"x": 1}
    cur.close()


def test_rownumber(conn):
    cur = conn.cursor("foo")
    assert cur.rownumber is None

    cur.execute("select 1 from generate_series(1, 42)")
    assert cur.rownumber == 0

    cur.fetchone()
    assert cur.rownumber == 1
    cur.fetchone()
    assert cur.rownumber == 2
    cur.fetchmany(10)
    assert cur.rownumber == 12
    cur.fetchall()
    assert cur.rownumber == 42
    cur.close()


def test_iter(conn):
    with conn.cursor("foo") as cur:
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (3,))
        recs = list(cur)
    assert recs == [(1,), (2,), (3,)]

    with conn.cursor("foo") as cur:
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (3,))
        assert cur.fetchone() == (1,)
        recs = list(cur)
    assert recs == [(2,), (3,)]


def test_iter_rownumber(conn):
    with conn.cursor("foo") as cur:
        cur.itersize = 2
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (3,))
        for row in cur:
            assert cur.rownumber == row[0]


def test_itersize(conn, commands):
    with conn.cursor("foo") as cur:
        assert cur.itersize == 100
        cur.itersize = 2
        cur.execute(ph(cur, "select generate_series(1, %s) as bar"), (3,))
        commands.popall()  # flush begin and other noise

        assert list(cur) == [(1,), (2,), (3,)]
        cmds = commands.popall()
        assert len(cmds) == 2
        for cmd in cmds:
            assert "fetch forward 2" in cmd.lower()


def test_next(conn):
    with conn.cursor() as cur:
        cur.execute("select 1")
        assert next(cur) == (1,)
        with pytest.raises(StopIteration):
            next(cur)


def test_cant_scroll_by_default(conn):
    cur = conn.cursor("tmp")
    assert cur.scrollable is None
    with pytest.raises(e.ProgrammingError):
        cur.scroll(0)
    cur.close()


@pytest.mark.crdb_skip("scroll cursor")
def test_scroll(conn):
    cur = conn.cursor("tmp", scrollable=True)
    cur.execute("select generate_series(0,9)")
    assert cur.rownumber == 0
    cur.scroll(2)
    assert cur.rownumber == 2
    assert cur.fetchone() == (2,)
    assert cur.rownumber == 3
    cur.scroll(2)
    assert cur.rownumber == 5
    assert cur.fetchone() == (5,)
    assert cur.rownumber == 6
    cur.scroll(2, mode="relative")
    assert cur.rownumber == 8
    assert cur.fetchone() == (8,)
    cur.scroll(9, mode="absolute")
    assert cur.rownumber == 9
    assert cur.fetchone() == (9,)
    assert cur.rownumber == 10

    with pytest.raises(ValueError):
        cur.scroll(9, mode="wat")
        assert cur.rownumber == 10
    cur.close()


@pytest.mark.crdb_skip("scroll cursor")
def test_scrollable(conn):
    curs = conn.cursor("foo", scrollable=True)
    assert curs.scrollable is True
    curs.execute("select generate_series(0, 5)")
    curs.scroll(5)
    for i in range(4, -1, -1):
        curs.scroll(-1)
        assert i == curs.fetchone()[0]
        curs.scroll(-1)
    curs.close()


def test_non_scrollable(conn):
    curs = conn.cursor("foo", scrollable=False)
    assert curs.scrollable is False
    curs.execute("select generate_series(0, 5)")
    curs.scroll(5)
    with pytest.raises(e.OperationalError):
        curs.scroll(-1)
    curs.close()


@pytest.mark.parametrize("kwargs", [{}, {"withhold": False}])
def test_no_hold(conn, kwargs):
    with conn.cursor("foo", **kwargs) as curs:
        assert curs.withhold is False
        curs.execute("select generate_series(0, 2)")
        assert curs.fetchone() == (0,)
        conn.commit()
        with pytest.raises(e.InvalidCursorName):
            curs.fetchone()


@pytest.mark.crdb_skip("cursor with hold")
def test_hold(conn):
    with conn.cursor("foo", withhold=True) as curs:
        assert curs.withhold is True
        curs.execute("select generate_series(0, 5)")
        assert curs.fetchone() == (0,)
        conn.commit()
        assert curs.fetchone() == (1,)


@pytest.mark.parametrize("row_factory", ["tuple_row", "namedtuple_row"])
def test_steal_cursor(conn, row_factory):
    cur1 = conn.cursor()
    cur1.execute("declare test cursor for select generate_series(1, 6) as s")

    cur2 = conn.cursor("test", row_factory=getattr(rows, row_factory))
    # can call fetch without execute
    rec = cur2.fetchone()
    assert rec == (1,)
    if row_factory == "namedtuple_row":
        assert rec.s == 1
    assert cur2.fetchmany(3) == [(2,), (3,), (4,)]
    assert cur2.fetchall() == [(5,), (6,)]
    cur2.close()


def test_stolen_cursor_close(conn):
    cur1 = conn.cursor()
    cur1.execute("declare test cursor for select generate_series(1, 6)")
    cur2 = conn.cursor("test")
    cur2.close()

    cur1.execute("declare test cursor for select generate_series(1, 6)")
    cur2 = conn.cursor("test")
    cur2.close()


def test_row_maker_returns_none(conn):
    query = "values (null), (0)"
    recs = [None, 0]
    with conn.cursor("test", row_factory=rows.scalar_row) as cur:
        cur.execute(query)
        assert [cur.fetchone() for _ in range(len(recs))] == recs
        cur.execute(query)
        assert cur.fetchmany(len(recs)) == recs
        cur.execute(query)
        assert cur.fetchall() == recs
        cur.execute(query)
        assert list(cur) == recs
        stream = cur.stream(query)
        assert list(stream) == recs


def test_results_after_execute(conn):
    with conn.cursor("test") as cur:
        cur.execute("select 1")
        assert list(cur.results()) == [cur]



# Include pid_t but Windows doesn't have it
# Don't use "IF" so that the generated C is portable and can be included
# in the sdist.
cdef extern from * nogil:
    """
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
    typedef signed pid_t;
#else
    #include <fcntl.h>
#endif
    """
    ctypedef signed pid_t

from psycopg_c.pq cimport libpq

ctypedef char *(*conn_bytes_f) (const libpq.PGconn *) noexcept nogil
ctypedef int (*conn_int_f) (const libpq.PGconn *) noexcept nogil
ctypedef void *(*conn_f_with_param) (const libpq.PGconn *, const char *) noexcept nogil
ctypedef int (*conn_int_f_with_param) (
    const libpq.PGconn *, const char *
) noexcept nogil


cdef class PGconn:
    cdef libpq.PGconn* _pgconn_ptr
    cdef object __weakref__
    cdef public object notice_handler
    cdef public object notify_handler
    cdef pid_t _procpid

    @staticmethod
    cdef PGconn _from_ptr(libpq.PGconn *ptr)

    cpdef int flush(self) except -1
    cpdef object notifies(self)


cdef class PGresult:
    cdef libpq.PGresult* _pgresult_ptr

    @staticmethod
    cdef PGresult _from_ptr(libpq.PGresult *ptr)


cdef class PGcancelConn:
    cdef libpq.PGcancelConn* pgcancelconn_ptr

    @staticmethod
    cdef PGcancelConn _from_ptr(libpq.PGcancelConn *ptr)


cdef class PGcancel:
    cdef libpq.PGcancel* pgcancel_ptr

    @staticmethod
    cdef PGcancel _from_ptr(libpq.PGcancel *ptr)


cdef class Escaping:
    cdef PGconn conn

    cpdef escape_literal(self, data)
    cpdef escape_identifier(self, data)
    cpdef escape_string(self, data)
    cpdef escape_bytea(self, data)
    cpdef unescape_bytea(self, const unsigned char *data)


cdef class PQBuffer:
    cdef unsigned char *buf
    cdef Py_ssize_t len

    @staticmethod
    cdef PQBuffer _from_buffer(unsigned char *buf, Py_ssize_t length)


cdef class ViewBuffer:
    cdef unsigned char *buf
    cdef Py_ssize_t len
    cdef object obj

    @staticmethod
    cdef ViewBuffer _from_buffer(
        object obj, unsigned char *buf, Py_ssize_t length)


cdef int _buffer_as_string_and_size(
    data: "Buffer", char **ptr, Py_ssize_t *length
) except -1



import pytest

import psycopg
from psycopg import pq

from ..fix_crdb import crdb_scs_off


@pytest.mark.parametrize(
    "data, want",
    [
        (b"", b"''"),
        (b"hello", b"'hello'"),
        (b"foo'bar", b"'foo''bar'"),
        (b"foo\\bar", b" E'foo\\\\bar'"),
    ],
)
def test_escape_literal(pgconn, data, want):
    esc = pq.Escaping(pgconn)
    out = esc.escape_literal(data)
    assert out == want


@pytest.mark.parametrize("scs", ["on", crdb_scs_off("off")])
def test_escape_literal_1char(pgconn, scs):
    res = pgconn.exec_(f"set standard_conforming_strings to {scs}".encode("ascii"))
    assert res.status == pq.ExecStatus.COMMAND_OK
    esc = pq.Escaping(pgconn)
    special = {b"'": b"''''", b"\\": b" E'\\\\'"}
    for c in range(1, 128):
        data = bytes([c])
        rv = esc.escape_literal(data)
        exp = special.get(data) or b"'%s'" % data
        assert rv == exp


def test_escape_literal_noconn(pgconn):
    esc = pq.Escaping()
    with pytest.raises(psycopg.OperationalError):
        esc.escape_literal(b"hi")

    esc = pq.Escaping(pgconn)
    pgconn.finish()
    with pytest.raises(psycopg.OperationalError):
        esc.escape_literal(b"hi")


@pytest.mark.parametrize(
    "data, want",
    [
        (b"", b'""'),
        (b"hello", b'"hello"'),
        (b'foo"bar', b'"foo""bar"'),
        (b"foo\\bar", b'"foo\\bar"'),
    ],
)
def test_escape_identifier(pgconn, data, want):
    esc = pq.Escaping(pgconn)
    out = esc.escape_identifier(data)
    assert out == want


@pytest.mark.parametrize("scs", ["on", crdb_scs_off("off")])
def test_escape_identifier_1char(pgconn, scs):
    res = pgconn.exec_(f"set standard_conforming_strings to {scs}".encode("ascii"))
    assert res.status == pq.ExecStatus.COMMAND_OK
    esc = pq.Escaping(pgconn)
    special = {b'"': b'""""', b"\\": b'"\\"'}
    for c in range(1, 128):
        data = bytes([c])
        rv = esc.escape_identifier(data)
        exp = special.get(data) or b'"%s"' % data
        assert rv == exp


def test_escape_identifier_noconn(pgconn):
    esc = pq.Escaping()
    with pytest.raises(psycopg.OperationalError):
        esc.escape_identifier(b"hi")

    esc = pq.Escaping(pgconn)
    pgconn.finish()
    with pytest.raises(psycopg.OperationalError):
        esc.escape_identifier(b"hi")


@pytest.mark.parametrize(
    "data, want",
    [
        (b"", b""),
        (b"hello", b"hello"),
        (b"foo'bar", b"foo''bar"),
        (b"foo\\bar", b"foo\\bar"),
    ],
)
def test_escape_string(pgconn, data, want):
    esc = pq.Escaping(pgconn)
    out = esc.escape_string(data)
    assert out == want


@pytest.mark.parametrize("scs", ["on", crdb_scs_off("off")])
def test_escape_string_1char(pgconn, scs):
    esc = pq.Escaping(pgconn)
    res = pgconn.exec_(f"set standard_conforming_strings to {scs}".encode("ascii"))
    assert res.status == pq.ExecStatus.COMMAND_OK
    special = {b"'": b"''", b"\\": b"\\" if scs == "on" else b"\\\\"}
    for c in range(1, 128):
        data = bytes([c])
        rv = esc.escape_string(data)
        exp = special.get(data) or b"%s" % data
        assert rv == exp


@pytest.mark.parametrize(
    "data, want",
    [
        (b"", b""),
        (b"hello", b"hello"),
        (b"foo'bar", b"foo''bar"),
        # This libpq function behaves unpredictably when not passed a conn
        (b"foo\\bar", (b"foo\\\\bar", b"foo\\bar")),
    ],
)
def test_escape_string_noconn(data, want):
    esc = pq.Escaping()
    out = esc.escape_string(data)
    if isinstance(want, bytes):
        assert out == want
    else:
        assert out in want


def test_escape_string_badconn(pgconn):
    esc = pq.Escaping(pgconn)
    pgconn.finish()
    with pytest.raises(psycopg.OperationalError):
        esc.escape_string(b"hi")


def test_escape_string_badenc(pgconn):
    res = pgconn.exec_(b"set client_encoding to 'UTF8'")
    assert res.status == pq.ExecStatus.COMMAND_OK
    data = "\u20ac".encode()[:-1]
    esc = pq.Escaping(pgconn)
    with pytest.raises(psycopg.OperationalError):
        esc.escape_string(data)


@pytest.mark.parametrize("data", [b"hello\00world", b"\00\00\00\00"])
def test_escape_bytea(pgconn, data):
    exp = rb"\x" + b"".join(b"%02x" % c for c in data)
    esc = pq.Escaping(pgconn)
    rv = esc.escape_bytea(data)
    assert rv == exp

    pgconn.finish()
    with pytest.raises(psycopg.OperationalError):
        esc.escape_bytea(data)


def test_escape_noconn(pgconn):
    data = bytes(range(256))
    esc = pq.Escaping()
    escdata = esc.escape_bytea(data)
    res = pgconn.exec_params(b"select '%s'::bytea" % escdata, [], result_format=1)
    assert res.status == pq.ExecStatus.TUPLES_OK
    assert res.get_value(0, 0) == data


def test_escape_1char(pgconn):
    esc = pq.Escaping(pgconn)
    for c in range(256):
        rv = esc.escape_bytea(bytes([c]))
        exp = rb"\x%02x" % c
        assert rv == exp


@pytest.mark.parametrize("data", [b"hello\00world", b"\00\00\00\00"])
def test_unescape_bytea(pgconn, data):
    enc = rb"\x" + b"".join(b"%02x" % c for c in data)
    esc = pq.Escaping(pgconn)
    rv = esc.unescape_bytea(enc)
    assert rv == data

    pgconn.finish()
    with pytest.raises(psycopg.OperationalError):
        esc.unescape_bytea(data)



# WARNING: this file is auto-generated by 'async_to_sync.py'
# from the original file 'test_sched_async.py'
# DO NOT CHANGE! Change the original file instead.
import logging
from time import time
from functools import partial
from contextlib import contextmanager

import pytest

from ..acompat import gather, sleep, spawn

try:
    from psycopg_pool.sched import Scheduler
except ImportError:
    # Tests should have been skipped if the package is not available
    pass

pytestmark = [pytest.mark.timing]


@pytest.mark.slow
def test_sched():
    s = Scheduler()
    results = []

    def worker(i):
        results.append((i, time()))

    t0 = time()
    s.enter(0.1, partial(worker, 1))
    s.enter(0.4, partial(worker, 3))
    s.enter(0.3, None)
    s.enter(0.2, partial(worker, 2))
    s.run()
    assert len(results) == 2
    assert results[0][0] == 1
    assert results[0][1] - t0 == pytest.approx(0.1, 0.1)
    assert results[1][0] == 2
    assert results[1][1] - t0 == pytest.approx(0.2, 0.1)


@pytest.mark.slow
def test_sched_task():
    s = Scheduler()
    t = spawn(s.run)

    results = []

    def worker(i):
        results.append((i, time()))

    t0 = time()
    s.enter(0.1, partial(worker, 1))
    s.enter(0.4, partial(worker, 3))
    s.enter(0.3, None)
    s.enter(0.2, partial(worker, 2))

    gather(t)
    t1 = time()
    assert t1 - t0 == pytest.approx(0.3, 0.2)

    assert len(results) == 2
    assert results[0][0] == 1
    assert results[0][1] - t0 == pytest.approx(0.1, 0.2)
    assert results[1][0] == 2
    assert results[1][1] - t0 == pytest.approx(0.2, 0.2)


@pytest.mark.slow
def test_sched_error(caplog):
    caplog.set_level(logging.WARNING, logger="psycopg")
    s = Scheduler()
    t = spawn(s.run)

    results = []

    def worker(i):
        results.append((i, time()))

    def error():
        1 / 0

    t0 = time()
    s.enter(0.1, partial(worker, 1))
    s.enter(0.4, None)
    s.enter(0.3, partial(worker, 2))
    s.enter(0.2, error)

    gather(t)
    t1 = time()
    assert t1 - t0 == pytest.approx(0.4, 0.1)

    assert len(results) == 2
    assert results[0][0] == 1
    assert results[0][1] - t0 == pytest.approx(0.1, 0.1)
    assert results[1][0] == 2
    assert results[1][1] - t0 == pytest.approx(0.3, 0.1)

    assert len(caplog.records) == 1
    assert "ZeroDivisionError" in caplog.records[0].message


@pytest.mark.slow
def test_empty_queue_timeout():
    s = Scheduler()

    with timed_wait(s) as times:
        s.EMPTY_QUEUE_TIMEOUT = 0.2

        t = spawn(s.run)
        sleep(0.5)
        s.enter(0.5, None)
        gather(t)

    for got, want in zip(times, [0.2, 0.4, 0.5, 1.0]):
        assert got == pytest.approx(want, 0.2), times


@pytest.mark.slow
def test_first_task_rescheduling():
    s = Scheduler()

    with timed_wait(s) as times:
        s.EMPTY_QUEUE_TIMEOUT = 0.1

        s.enter(0.4, noop)
        t = spawn(s.run)
        s.enter(0.6, None)  # this task doesn't trigger a reschedule
        sleep(0.1)
        s.enter(0.1, noop)  # this triggers a reschedule
        gather(t)

    for got, want in zip(times, [0.1, 0.2, 0.4, 0.6, 0.6]):
        assert got == pytest.approx(want, 0.2), times


@contextmanager
def timed_wait(s):
    """
    Hack the scheduler's Event.wait() function in order to log waited time.

    The context is a list where the times are accumulated.
    """
    t0 = time()
    times = []

    wait_orig = s._event.wait

    def wait_logging(timeout=None):
        args = (timeout,)

        try:
            rv = wait_orig(*args)
        finally:
            times.append(time() - t0)
        return rv

    setattr(s._event, "wait", wait_logging)

    yield times

    times.append(time() - t0)


def noop():
    pass



"""
psycopg_c.pq.PGconn object implementation.
"""

# Copyright (C) 2020 The Psycopg Team

cdef extern from * nogil:
    """
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
    /* We don't need a real definition for this because Windows is not affected
     * by the issue caused by closing the fds after fork.
     */
    #define getpid() (0)
#else
    #include <unistd.h>
#endif
    """
    pid_t getpid()

cimport cython
from libc.stdio cimport fdopen
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from cpython.bytes cimport PyBytes_AsString
from cpython.object cimport PyObject
from cpython.sequence cimport PySequence_Fast, PySequence_Fast_GET_ITEM
from cpython.sequence cimport PySequence_Fast_GET_SIZE
from cpython.memoryview cimport PyMemoryView_FromObject

import sys

from psycopg.pq import Format as PqFormat
from psycopg.pq import Trace, version_pretty
from psycopg.pq.misc import PGnotify, _clean_error_message, connection_summary
from psycopg.pq._enums import ExecStatus
from psycopg._encodings import pg2pyenc

from psycopg_c.pq cimport PQBuffer


cdef object _check_supported(fname, int pgversion):
    if libpq.PG_VERSION_NUM < pgversion:
        raise e.NotSupportedError(
            f"{fname} requires libpq from PostgreSQL {version_pretty(pgversion)}"
            f" on the client; version {version_pretty(libpq.PG_VERSION_NUM)}"
            " available instead"
        )

cdef class PGconn:
    @staticmethod
    cdef PGconn _from_ptr(libpq.PGconn *ptr):
        cdef PGconn rv = PGconn.__new__(PGconn)
        rv._pgconn_ptr = ptr

        libpq.PQsetNoticeReceiver(
            ptr, <libpq.PQnoticeReceiver>notice_receiver, <void *>rv)
        return rv

    def __cinit__(self):
        self._pgconn_ptr = NULL
        self._procpid = getpid()

    def __dealloc__(self):
        # Close the connection only if it was created in this process,
        # not if this object is being GC'd after fork.
        if self._procpid == getpid():
            self.finish()

    def __repr__(self) -> str:
        cls = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        info = connection_summary(self)
        return f"<{cls} {info} at 0x{id(self):x}>"

    @classmethod
    def connect(cls, const char *conninfo) -> PGconn:
        cdef libpq.PGconn* pgconn = libpq.PQconnectdb(conninfo)
        if not pgconn:
            raise MemoryError("couldn't allocate PGconn")

        return PGconn._from_ptr(pgconn)

    @classmethod
    def connect_start(cls, const char *conninfo) -> PGconn:
        cdef libpq.PGconn* pgconn = libpq.PQconnectStart(conninfo)
        if not pgconn:
            raise MemoryError("couldn't allocate PGconn")

        return PGconn._from_ptr(pgconn)

    def connect_poll(self) -> int:
        return _call_libpq_int(self, <conn_int_f>libpq.PQconnectPoll)

    def finish(self) -> None:
        with cython.critical_section(self):
            if self._pgconn_ptr is not NULL:
                libpq.PQfinish(self._pgconn_ptr)
                self._pgconn_ptr = NULL

    @property
    def pgconn_ptr(self) -> int | None:
        cdef long long ptr = -1
        with cython.critical_section(self):
            if self._pgconn_ptr:
                ptr = <long long><void *>self._pgconn_ptr
        return ptr if ptr != -1 else None

    @property
    def info(self) -> list[ConninfoOption]:
        cdef libpq.PQconninfoOption *opts
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                opts = libpq.PQconninfo(pgconn_ptr)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if opts is NULL:
            raise MemoryError("couldn't allocate connection info")
        rv = _options_from_array(opts)
        libpq.PQconninfoFree(opts)
        return rv

    def reset(self) -> None:
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                libpq.PQreset(pgconn_ptr)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")

    def reset_start(self) -> None:
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQresetStart(self._pgconn_ptr)
        if not rv:
            raise e.OperationalError("couldn't reset connection")

    def reset_poll(self) -> int:
        return _call_libpq_int(self, <conn_int_f>libpq.PQresetPoll)

    @classmethod
    def ping(self, const char *conninfo) -> int:
        return libpq.PQping(conninfo)

    @property
    def db(self) -> bytes:
        return _call_libpq_bytes(self, libpq.PQdb)

    @property
    def user(self) -> bytes:
        return _call_libpq_bytes(self, libpq.PQuser)

    @property
    def password(self) -> bytes:
        return _call_libpq_bytes(self, libpq.PQpass)

    @property
    def host(self) -> bytes:
        return _call_libpq_bytes(self, libpq.PQhost)

    @property
    def hostaddr(self) -> bytes:
        _check_supported("PQhostaddr", 120000)
        cdef char *rv = NULL
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQhostaddr(pgconn_ptr)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        assert rv is not NULL
        return rv

    @property
    def port(self) -> bytes:
        return _call_libpq_bytes(self, libpq.PQport)

    @property
    def tty(self) -> bytes:
        return _call_libpq_bytes(self, libpq.PQtty)

    @property
    def options(self) -> bytes:
        return _call_libpq_bytes(self, libpq.PQoptions)

    @property
    def status(self) -> int:
        cdef libpq.ConnStatusType rv
        with cython.critical_section(self):
            rv = libpq.PQstatus(self._pgconn_ptr)
        return rv

    @property
    def transaction_status(self) -> int:
        cdef libpq.PGTransactionStatusType rv
        with cython.critical_section(self):
            rv = libpq.PQtransactionStatus(self._pgconn_ptr)
        return rv

    def parameter_status(self, const char *name) -> bytes | None:
        cdef const char *rv = <const char *>_call_libpq_with_param(
            self, <conn_f_with_param>libpq.PQparameterStatus, name)
        if rv is not NULL:
            return rv
        else:
            return None

    @property
    def error_message(self) -> bytes:
        cdef char *rv
        with cython.critical_section(self):
            rv = libpq.PQerrorMessage(self._pgconn_ptr)
        return rv

    def get_error_message(self, encoding: str = "") -> str:
        return _clean_error_message(self.error_message, encoding or self._encoding)

    @property
    def _encoding(self) -> str:
        cdef const char *pgenc
        cdef int status
        with cython.critical_section(self):
            status = libpq.PQstatus(self._pgconn_ptr)
            if status == libpq.CONNECTION_OK:
                pgenc = libpq.PQparameterStatus(self._pgconn_ptr, b"client_encoding")
                if pgenc is NULL:
                    pgenc = b"UTF8"
        if status == libpq.CONNECTION_OK:
            return pg2pyenc(pgenc)
        else:
            return "utf-8"

    @property
    def protocol_version(self) -> int:
        return _call_libpq_int(self, <conn_int_f>libpq.PQprotocolVersion)

    @property
    def full_protocol_version(self) -> int:
        _check_supported("PQfullProtocolVersion", 180000)
        cdef int rv
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQfullProtocolVersion(pgconn_ptr)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        return rv

    @property
    def server_version(self) -> int:
        return _call_libpq_int(self, <conn_int_f>libpq.PQserverVersion)

    @property
    def socket(self) -> int:
        cdef int rv = _call_libpq_int(self, <conn_int_f>libpq.PQsocket)
        if rv == -1:
            raise e.OperationalError("the connection is lost")
        return rv

    @property
    def backend_pid(self) -> int:
        return _call_libpq_int(self, <conn_int_f>libpq.PQbackendPID)

    @property
    def needs_password(self) -> bool:
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQconnectionNeedsPassword(self._pgconn_ptr)
        return bool(rv)

    @property
    def used_password(self) -> bool:
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQconnectionUsedPassword(self._pgconn_ptr)
        return bool(rv)

    @property
    def used_gssapi(self) -> bool:
        _check_supported("PQconnectionUsedGSSAPI", 160000)
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQconnectionUsedGSSAPI(self._pgconn_ptr)
        return bool(rv)

    @property
    def ssl_in_use(self) -> bool:
        return bool(_call_libpq_int(self, <conn_int_f>libpq.PQsslInUse))

    def exec_(self, const char *command) -> PGresult:
        cdef libpq.PGresult *pgresult
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self), nogil:
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                pgresult = libpq.PQexec(pgconn_ptr, command)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if pgresult is NULL:
            raise e.OperationalError(
                f"executing query failed: {self.get_error_message()}")
        return PGresult._from_ptr(pgresult)

    def send_query(self, const char *command) -> None:
        cdef int rv
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self), nogil:
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQsendQuery(pgconn_ptr, command)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if not rv:
            raise e.OperationalError(
                f"sending query failed: {self.get_error_message()}")

    def exec_params(
        self,
        const char *command,
        param_values: Sequence[bytes | None] | None,
        param_types: Sequence[int] | None = None,
        param_formats: Sequence[int] | None = None,
        int result_format = PqFormat.TEXT,
    ) -> PGresult:
        cdef Py_ssize_t cnparams
        cdef libpq.Oid *ctypes
        cdef char *const *cvalues
        cdef int *clengths
        cdef int *cformats
        cnparams, ctypes, cvalues, clengths, cformats = _query_params_args(
            param_values, param_types, param_formats)

        cdef libpq.PGconn *pgconn_ptr
        cdef libpq.PGresult *pgresult
        with cython.critical_section(self), nogil:
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                pgresult = libpq.PQexecParams(
                    pgconn_ptr, command, <int>cnparams, ctypes,
                    <const char *const *>cvalues, clengths, cformats, result_format)
        _clear_query_params(ctypes, cvalues, clengths, cformats)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if pgresult is NULL:
            raise e.OperationalError(
                f"executing query failed: {self.get_error_message()}")
        return PGresult._from_ptr(pgresult)

    def send_query_params(
        self,
        const char *command,
        param_values: Sequence[bytes | None] | None,
        param_types: Sequence[int] | None = None,
        param_formats: Sequence[int] | None = None,
        int result_format = PqFormat.TEXT,
    ) -> None:
        cdef Py_ssize_t cnparams
        cdef libpq.Oid *ctypes
        cdef char *const *cvalues
        cdef int *clengths
        cdef int *cformats
        cnparams, ctypes, cvalues, clengths, cformats = _query_params_args(
            param_values, param_types, param_formats)

        cdef libpq.PGconn *pgconn_ptr
        cdef int rv
        with cython.critical_section(self), nogil:
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQsendQueryParams(
                    pgconn_ptr, command, <int>cnparams, ctypes,
                    <const char *const *>cvalues, clengths, cformats, result_format)
        _clear_query_params(ctypes, cvalues, clengths, cformats)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if not rv:
            raise e.OperationalError(
                f"sending query and params failed: {self.get_error_message()}"
            )

    def send_prepare(
        self,
        const char *name,
        const char *command,
        param_types: Sequence[int] | None = None,
    ) -> None:
        cdef int i
        cdef types_fast
        cdef Py_ssize_t nparams = 0
        if param_types is not None:
            types_fast = PySequence_Fast(param_types, "'param_types' is not a sequence")
            nparams = PySequence_Fast_GET_SIZE(types_fast)

        cdef libpq.Oid *atypes = NULL
        if nparams:
            atypes = <libpq.Oid *>PyMem_Malloc(nparams * sizeof(libpq.Oid))
            for i in range(nparams):
                atypes[i] = <object>PySequence_Fast_GET_ITEM(types_fast, i)

        cdef libpq.PGconn *pgconn_ptr
        cdef int rv
        with cython.critical_section(self), nogil:
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQsendPrepare(
                    pgconn_ptr, name, command, <int>nparams, atypes
                )
        PyMem_Free(atypes)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if not rv:
            raise e.OperationalError(
                f"sending query and params failed: {self.get_error_message()}"
            )

    def send_query_prepared(
        self,
        const char *name,
        param_values: Sequence[bytes | None] | None,
        param_formats: Sequence[int] | None = None,
        int result_format = PqFormat.TEXT,
    ) -> None:
        cdef Py_ssize_t cnparams
        cdef libpq.Oid *ctypes
        cdef char *const *cvalues
        cdef int *clengths
        cdef int *cformats
        cnparams, ctypes, cvalues, clengths, cformats = _query_params_args(
            param_values, None, param_formats)

        cdef libpq.PGconn *pgconn_ptr
        cdef int rv
        with cython.critical_section(self), nogil:
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQsendQueryPrepared(
                    pgconn_ptr, name, <int>cnparams, <const char *const *>cvalues,
                    clengths, cformats, result_format)
        _clear_query_params(ctypes, cvalues, clengths, cformats)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if not rv:
            raise e.OperationalError(
                f"sending prepared query failed: {self.get_error_message()}"
            )

    def prepare(
        self,
        const char *name,
        const char *command,
        param_types: Sequence[int] | None = None,
    ) -> PGresult:
        cdef int i
        cdef types_fast
        cdef Py_ssize_t nparams = 0
        if param_types is not None:
            types_fast = PySequence_Fast(param_types, "'param_types' is not a sequence")
            nparams = PySequence_Fast_GET_SIZE(types_fast)

        cdef libpq.Oid *atypes = NULL
        if nparams:
            atypes = <libpq.Oid *>PyMem_Malloc(nparams * sizeof(libpq.Oid))
            for i in range(nparams):
                atypes[i] = <object>PySequence_Fast_GET_ITEM(types_fast, i)

        cdef libpq.PGconn *pgconn_ptr
        cdef libpq.PGresult *rv
        with cython.critical_section(self), nogil:
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQprepare(
                    pgconn_ptr, name, command, <int>nparams, atypes)
        PyMem_Free(atypes)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if rv is NULL:
            raise e.OperationalError(
                f"preparing query failed: {self.get_error_message()}")
        return PGresult._from_ptr(rv)

    def exec_prepared(
        self,
        const char *name,
        param_values: Sequence[bytes] | None,
        param_formats: Sequence[int] | None = None,
        int result_format = PqFormat.TEXT,
    ) -> PGresult:
        cdef Py_ssize_t cnparams
        cdef libpq.Oid *ctypes
        cdef char *const *cvalues
        cdef int *clengths
        cdef int *cformats
        cnparams, ctypes, cvalues, clengths, cformats = _query_params_args(
            param_values, None, param_formats)

        cdef libpq.PGconn *pgconn_ptr
        cdef libpq.PGresult *rv
        with cython.critical_section(self), nogil:
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQexecPrepared(
                    pgconn_ptr, name, <int>cnparams,
                    <const char *const *>cvalues,
                    clengths, cformats, result_format)
        _clear_query_params(ctypes, cvalues, clengths, cformats)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if rv is NULL:
            raise e.OperationalError(
                f"executing prepared query failed: {self.get_error_message()}"
            )
        return PGresult._from_ptr(rv)

    def describe_prepared(self, const char *name) -> PGresult:
        cdef libpq.PGresult *rv = <libpq.PGresult *>_call_libpq_with_param(
            self, <conn_f_with_param>libpq.PQdescribePrepared, name)
        if rv is NULL:
            raise e.OperationalError(
                f"describe prepared failed: {self.get_error_message()}"
            )
        return PGresult._from_ptr(rv)

    def send_describe_prepared(self, const char *name) -> None:
        cdef int rv = _call_libpq_int_with_param(
            self, <conn_int_f_with_param>libpq.PQsendDescribePrepared, name)
        if not rv:
            raise e.OperationalError(
                f"sending describe prepared failed: {self.get_error_message()}"
            )

    def describe_portal(self, const char *name) -> PGresult:
        cdef libpq.PGresult *rv = <libpq.PGresult *>_call_libpq_with_param(
            self, <conn_f_with_param>libpq.PQdescribePortal, name)
        if rv is NULL:
            raise e.OperationalError(
                f"describe prepared failed: {self.get_error_message()}"
            )
        return PGresult._from_ptr(rv)

    def send_describe_portal(self, const char *name) -> None:
        cdef int rv = _call_libpq_int_with_param(
            self, <conn_int_f_with_param>libpq.PQsendDescribePortal, name)
        if not rv:
            raise e.OperationalError(
                f"sending describe prepared failed: {self.get_error_message()}"
            )

    def close_prepared(self, const char *name) -> PGresult:
        _check_supported("PQclosePrepared", 170000)
        cdef libpq.PGresult *rv
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQclosePrepared(pgconn_ptr, name)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if rv is NULL:
            raise e.OperationalError(
                f"close prepared failed: {self.get_error_message()}"
            )
        return PGresult._from_ptr(rv)

    def send_close_prepared(self, const char *name) -> None:
        _check_supported("PQsendClosePrepared", 170000)
        cdef int rv
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQsendClosePrepared(pgconn_ptr, name)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if not rv:
            raise e.OperationalError(
                f"sending close prepared failed: {self.get_error_message()}"
            )

    def close_portal(self, const char *name) -> PGresult:
        _check_supported("PQclosePortal", 170000)
        cdef libpq.PGresult *rv
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQclosePortal(pgconn_ptr, name)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if rv is NULL:
            raise e.OperationalError(
                f"close prepared failed: {self.get_error_message()}"
            )
        return PGresult._from_ptr(rv)

    def send_close_portal(self, const char *name) -> None:
        _check_supported("PQsendClosePortal", 170000)
        cdef int rv
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQsendClosePortal(pgconn_ptr, name)
        if pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")
        if not rv:
            raise e.OperationalError(
                f"sending close prepared failed: {self.get_error_message()}"
            )

    def get_result(self) -> "PGresult" | None:
        cdef libpq.PGresult *pgresult
        with cython.critical_section(self):
            pgresult = libpq.PQgetResult(self._pgconn_ptr)
        if pgresult is NULL:
            return None
        return PGresult._from_ptr(pgresult)

    def consume_input(self) -> None:
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQconsumeInput(self._pgconn_ptr)
        if 1 != rv:
            raise e.OperationalError(
                f"consuming input failed: {self.get_error_message()}")

    def is_busy(self) -> int:
        cdef int rv
        with cython.critical_section(self), nogil:
            rv = libpq.PQisBusy(self._pgconn_ptr)
        return rv

    @property
    def nonblocking(self) -> int:
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQisnonblocking(self._pgconn_ptr)
        return rv

    @nonblocking.setter
    def nonblocking(self, int arg) -> None:
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQsetnonblocking(self._pgconn_ptr, arg)
        if 0 > rv:
            raise e.OperationalError(
                f"setting nonblocking failed: {self.get_error_message()}")

    cpdef int flush(self) except -1:
        cdef int rv
        cdef libpq.PGconn *pgconn_ptr
        with cython.critical_section(self):
            if (pgconn_ptr := self._pgconn_ptr) is not NULL:
                rv = libpq.PQflush(self._pgconn_ptr)
        if pgconn_ptr is NULL:
            raise e.OperationalError("flushing failed: the connection is closed")
        if rv < 0:
            raise e.OperationalError(f"flushing failed: {self.get_error_message()}")
        return rv

    def set_single_row_mode(self) -> None:
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQsetSingleRowMode(self._pgconn_ptr)
        if not rv:
            raise e.OperationalError("setting single row mode failed")

    def set_chunked_rows_mode(self, size: int) -> None:
        cdef int rv
        cdef int s = size
        with cython.critical_section(self):
            rv = libpq.PQsetChunkedRowsMode(self._pgconn_ptr, s)
        if not rv:
            raise e.OperationalError("setting chunked rows mode failed")

    def cancel_conn(self) -> PGcancelConn:
        _check_supported("PQcancelCreate", 170000)
        cdef libpq.PGcancelConn *ptr
        with cython.critical_section(self):
            ptr = libpq.PQcancelCreate(self._pgconn_ptr)
        if not ptr:
            raise e.OperationalError("couldn't create cancelConn object")
        return PGcancelConn._from_ptr(ptr)

    def get_cancel(self) -> PGcancel:
        cdef libpq.PGcancel *ptr
        with cython.critical_section(self):
            ptr = libpq.PQgetCancel(self._pgconn_ptr)
        if not ptr:
            raise e.OperationalError("couldn't create cancel object")
        return PGcancel._from_ptr(ptr)

    cpdef object notifies(self):
        cdef libpq.PGnotify *ptr
        with cython.critical_section(self), nogil:
            ptr = libpq.PQnotifies(self._pgconn_ptr)
        if ptr:
            ret = PGnotify(ptr.relname, ptr.be_pid, ptr.extra)
            libpq.PQfreemem(ptr)
            return ret
        else:
            return None

    def put_copy_data(self, buffer) -> int:
        cdef int rv
        cdef char *cbuffer
        cdef Py_ssize_t length

        _buffer_as_string_and_size(buffer, &cbuffer, &length)
        with cython.critical_section(self):
            rv = libpq.PQputCopyData(self._pgconn_ptr, cbuffer, <int>length)
        if rv < 0:
            raise e.OperationalError(
                f"sending copy data failed: {self.get_error_message()}")
        return rv

    def put_copy_end(self, error: bytes | None = None) -> int:
        cdef int rv
        cdef const char *cerr = NULL
        if error is not None:
            cerr = PyBytes_AsString(error)

        with cython.critical_section(self):
            rv = libpq.PQputCopyEnd(self._pgconn_ptr, cerr)
        if rv < 0:
            raise e.OperationalError(
                f"sending copy end failed: {self.get_error_message()}")
        return rv

    def get_copy_data(self, int async_) -> tuple[int, memoryview]:
        cdef char *buffer_ptr = NULL
        cdef int nbytes

        with cython.critical_section(self):
            nbytes = libpq.PQgetCopyData(self._pgconn_ptr, &buffer_ptr, async_)
        if nbytes == -2:
            raise e.OperationalError(
                f"receiving copy data failed: {self.get_error_message()}")
        if buffer_ptr is not NULL:
            data = PyMemoryView_FromObject(
                PQBuffer._from_buffer(<unsigned char *>buffer_ptr, nbytes))
            return nbytes, data
        else:
            return nbytes, b""  # won't parse it, doesn't really be memoryview

    def trace(self, fileno: int) -> None:
        if sys.platform != "linux":
            raise e.NotSupportedError("currently only supported on Linux")
        stream = fdopen(fileno, b"w")
        with cython.critical_section(self):
            libpq.PQtrace(self._pgconn_ptr, stream)

    def set_trace_flags(self, flags: Trace) -> None:
        _check_supported("PQsetTraceFlags", 140000)
        cdef int f = flags
        with cython.critical_section(self):
            libpq.PQsetTraceFlags(self._pgconn_ptr, f)

    def untrace(self) -> None:
        with cython.critical_section(self):
            libpq.PQuntrace(self._pgconn_ptr)

    def encrypt_password(
        self, const char *passwd, const char *user, algorithm = None
    ) -> bytes:
        _check_supported("PQencryptPasswordConn", 100000)

        cdef char *out
        cdef const char *calgo = NULL
        if algorithm:
            calgo = algorithm
        with cython.critical_section(self):
            out = libpq.PQencryptPasswordConn(self._pgconn_ptr, passwd, user, calgo)
        if not out:
            raise e.OperationalError(
                f"password encryption failed: {self.get_error_message()}"
            )

        rv = bytes(out)
        libpq.PQfreemem(out)
        return rv

    def change_password(
        self, const char *user, const char *passwd
    ) -> None:
        _check_supported("PQchangePassword", 170000)

        cdef libpq.PGresult *res
        with cython.critical_section(self):
            res = libpq.PQchangePassword(self._pgconn_ptr, user, passwd)
        if libpq.PQresultStatus(res) != ExecStatus.COMMAND_OK:
            raise e.OperationalError(
                f"password encryption failed: {self.get_error_message()}"
            )

    def make_empty_result(self, int exec_status) -> PGresult:
        cdef libpq.PGresult *rv
        with cython.critical_section(self):
            rv = libpq.PQmakeEmptyPGresult(
                self._pgconn_ptr, <libpq.ExecStatusType>exec_status)
        if not rv:
            raise MemoryError("couldn't allocate empty PGresult")
        return PGresult._from_ptr(rv)

    @property
    def pipeline_status(self) -> int:
        """The current pipeline mode status.

        For libpq < 14.0, always return 0 (PQ_PIPELINE_OFF).
        """
        if libpq.PG_VERSION_NUM < 140000:
            return libpq.PQ_PIPELINE_OFF
        cdef libpq.PGpipelineStatus rv
        with cython.critical_section(self):
            rv = libpq.PQpipelineStatus(self._pgconn_ptr)
        return rv

    def enter_pipeline_mode(self) -> None:
        """Enter pipeline mode.

        :raises ~e.OperationalError: in case of failure to enter the pipeline
            mode.
        """
        _check_supported("PQenterPipelineMode", 140000)
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQenterPipelineMode(self._pgconn_ptr)
        if rv != 1:
            raise e.OperationalError("failed to enter pipeline mode")

    def exit_pipeline_mode(self) -> None:
        """Exit pipeline mode.

        :raises ~e.OperationalError: in case of failure to exit the pipeline
            mode.
        """
        _check_supported("PQexitPipelineMode", 140000)
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQexitPipelineMode(self._pgconn_ptr)
        if rv != 1:
            raise e.OperationalError(self.get_error_message())

    def pipeline_sync(self) -> None:
        """Mark a synchronization point in a pipeline.

        :raises ~e.OperationalError: if the connection is not in pipeline mode
            or if sync failed.
        """
        _check_supported("PQpipelineSync", 140000)
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQpipelineSync(self._pgconn_ptr)
        if rv == 0:
            raise e.OperationalError("connection not in pipeline mode")
        if rv != 1:
            raise e.OperationalError("failed to sync pipeline")

    def send_flush_request(self) -> None:
        """Sends a request for the server to flush its output buffer.

        :raises ~e.OperationalError: if the flush request failed.
        """
        _check_supported("PQsendFlushRequest ", 140000)
        cdef int rv
        with cython.critical_section(self):
            rv = libpq.PQsendFlushRequest(self._pgconn_ptr)
        if rv == 0:
            raise e.OperationalError(
                f"flush request failed: {self.get_error_message()}")


cdef void *_call_libpq_with_param(
    PGconn self,
    conn_f_with_param func,
    const char *param
):
    cdef void *rv = NULL
    cdef libpq.PGconn *pgconn_ptr
    with cython.critical_section(self):
        if (pgconn_ptr := self._pgconn_ptr) is not NULL:
            rv = func(pgconn_ptr, param)
    if pgconn_ptr is NULL:
        raise e.OperationalError("the connection is closed")
    return rv


cdef int _call_libpq_int(PGconn self, conn_int_f func):
    cdef int rv
    cdef libpq.PGconn *pgconn_ptr
    with cython.critical_section(self):
        if (pgconn_ptr := self._pgconn_ptr) is not NULL:
            rv = func(pgconn_ptr)
    if pgconn_ptr is NULL:
        raise e.OperationalError("the connection is closed")
    return rv


cdef int _call_libpq_int_with_param(
    PGconn self,
    conn_int_f_with_param func,
    const char *param
):
    cdef int rv
    cdef libpq.PGconn *pgconn_ptr
    with cython.critical_section(self):
        if (pgconn_ptr := self._pgconn_ptr) is not NULL:
            rv = func(pgconn_ptr, param)
    if pgconn_ptr is NULL:
        raise e.OperationalError("the connection is closed")
    return rv


cdef bytes _call_libpq_bytes(PGconn self, conn_bytes_f func):
    cdef char *rv
    cdef libpq.PGconn *pgconn_ptr
    with cython.critical_section(self):
        if (pgconn_ptr := self._pgconn_ptr) is not NULL:
            rv = func(pgconn_ptr)
    if pgconn_ptr is NULL:
        raise e.OperationalError("the connection is closed")
    if rv is not NULL:
        return rv
    return b""


cdef void notice_receiver(void *arg, const libpq.PGresult *res_ptr) noexcept with gil:
    cdef PGconn pgconn = <object>arg
    if pgconn.notice_handler is None:
        return

    cdef PGresult res = PGresult._from_ptr(<libpq.PGresult *>res_ptr)
    try:
        pgconn.notice_handler(res)
    except Exception as e:
        logger.exception("error in notice receiver: %s", e)
    finally:
        res._pgresult_ptr = NULL  # avoid destroying the pgresult_ptr


cdef (Py_ssize_t, libpq.Oid *, char * const*, int *, int *) _query_params_args(
    param_values: Sequence[bytes | None] | None,
    param_types: Sequence[int] | None,
    param_formats: Sequence[int] | None,
) except *:
    cdef int i

    cdef values_fast
    cdef types_fast
    cdef formats_fast

    cdef Py_ssize_t nparams = 0
    if param_values is not None:
        values_fast = PySequence_Fast(param_values, "'param_values' is not a sequence")
        nparams = PySequence_Fast_GET_SIZE(values_fast)

    if param_types is not None:
        types_fast = PySequence_Fast(param_types, "'param_types' is not a sequence")
        if PySequence_Fast_GET_SIZE(types_fast) != nparams:
            raise ValueError(
                f"got {nparams} param_values but {len(param_types)} param_types"
            )

    if param_formats is not None:
        formats_fast = PySequence_Fast(
            param_formats, "'param_formats' is not a sequence")
        if PySequence_Fast_GET_SIZE(formats_fast) != nparams:
            raise ValueError(
                f"got {nparams} param_values but {len(param_formats)} param_formats"
            )

    cdef char **aparams = NULL
    cdef int *alenghts = NULL
    cdef char *ptr
    cdef Py_ssize_t length

    if nparams:
        aparams = <char **>PyMem_Malloc(nparams * sizeof(char *))
        alenghts = <int *>PyMem_Malloc(nparams * sizeof(int))
        for i in range(nparams):
            obj = PySequence_Fast_GET_ITEM(values_fast, i)
            if obj != <PyObject *>None:
                # TODO: it is a leak if this fails (but it should only fail
                # on internal error, e.g. if obj is not a buffer)
                _buffer_as_string_and_size(<object>obj, &ptr, &length)
                aparams[i] = ptr
                alenghts[i] = <int>length
            else:
                aparams[i] = NULL
                alenghts[i] = 0

    cdef libpq.Oid *atypes = NULL
    if param_types is not None:
        atypes = <libpq.Oid *>PyMem_Malloc(nparams * sizeof(libpq.Oid))
        for i in range(nparams):
            atypes[i] = <object>PySequence_Fast_GET_ITEM(types_fast, i)

    cdef int *aformats = NULL
    if param_formats is not None:
        aformats = <int *>PyMem_Malloc(nparams * sizeof(int *))
        for i in range(nparams):
            aformats[i] = <object>PySequence_Fast_GET_ITEM(formats_fast, i)

    return (nparams, atypes, aparams, alenghts, aformats)


cdef void _clear_query_params(
    libpq.Oid *ctypes, char *const *cvalues, int *clenghst, int *cformats
):
    PyMem_Free(ctypes)
    PyMem_Free(<char **>cvalues)
    PyMem_Free(clenghst)
    PyMem_Free(cformats)



"""
Prepared statements tests
"""

import sys
import logging
import datetime as dt
from decimal import Decimal

import pytest

import psycopg
from psycopg.rows import namedtuple_row
from psycopg.pq._debug import PGconnDebug


@pytest.mark.parametrize("value", [None, 0, 3])
async def test_prepare_threshold_init(aconn_cls, dsn, value):
    async with await aconn_cls.connect(dsn, prepare_threshold=value) as conn:
        assert conn.prepare_threshold == value


async def test_dont_prepare(aconn):
    cur = aconn.cursor()
    for i in range(10):
        await cur.execute("select %s::int", [i], prepare=False)

    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 0


async def test_do_prepare(aconn):
    cur = aconn.cursor()
    await cur.execute("select %s::int", [10], prepare=True)
    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 1


async def test_auto_prepare(aconn):
    res = []
    for i in range(10):
        await aconn.execute("select %s::int", [0])
        stmts = await get_prepared_statements(aconn)
        res.append(len(stmts))

    assert res == [0] * 5 + [1] * 5


async def test_dont_prepare_conn(aconn):
    for i in range(10):
        await aconn.execute("select %s::int", [i], prepare=False)

    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 0


async def test_do_prepare_conn(aconn):
    await aconn.execute("select %s::int", [10], prepare=True)
    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 1


async def test_auto_prepare_conn(aconn):
    res = []
    for i in range(10):
        await aconn.execute("select %s", [0])
        stmts = await get_prepared_statements(aconn)
        res.append(len(stmts))

    assert res == [0] * 5 + [1] * 5


async def test_prepare_disable(aconn):
    aconn.prepare_threshold = None
    res = []
    for i in range(10):
        await aconn.execute("select %s", [0])
        stmts = await get_prepared_statements(aconn)
        res.append(len(stmts))

    assert res == [0] * 10
    assert not aconn._prepared._names
    assert not aconn._prepared._counts


async def test_no_prepare_multi(aconn):
    res = []
    for i in range(10):
        await aconn.execute("select 1; select 2")
        stmts = await get_prepared_statements(aconn)
        res.append(len(stmts))

    assert res == [0] * 10


async def test_no_prepare_multi_with_drop(aconn):
    await aconn.execute("select 1", prepare=True)

    for i in range(10):
        await aconn.execute("drop table if exists noprep; create table noprep()")

    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 0


async def test_no_prepare_error(aconn):
    await aconn.set_autocommit(True)
    for i in range(10):
        with pytest.raises(aconn.ProgrammingError):
            await aconn.execute("select wat")

    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 0


@pytest.mark.parametrize(
    "query",
    [
        "create table test_no_prepare ()",
        pytest.param("notify foo, 'bar'", marks=pytest.mark.crdb_skip("notify")),
        "set timezone = utc",
        "select num from prepared_test",
        "insert into prepared_test (num) values (1)",
        "update prepared_test set num = num * 2",
        "delete from prepared_test where num > 10",
    ],
)
async def test_misc_statement(aconn, query):
    await aconn.execute("create table prepared_test (num int)", prepare=False)
    aconn.prepare_threshold = 0
    await aconn.execute(query)
    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 1


async def test_params_types(aconn):
    await aconn.execute(
        "select %s, %s, %s",
        [dt.date(2020, 12, 10), 42, Decimal(42)],
        prepare=True,
    )
    stmts = await get_prepared_statements(aconn)
    want = [stmt.parameter_types for stmt in stmts]
    assert want == [["date", "smallint", "numeric"]]


async def test_evict_lru(aconn):
    aconn.prepared_max = 5
    for i in range(10):
        await aconn.execute("select 'a'")
        await aconn.execute(f"select {i}")

    assert len(aconn._prepared._names) == 1
    assert aconn._prepared._names[b"select 'a'", ()] == b"_pg3_0"
    for i in [9, 8, 7, 6]:
        assert aconn._prepared._counts[f"select {i}".encode(), ()] == 1

    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 1
    assert stmts[0].statement == "select 'a'"


async def test_evict_lru_deallocate(aconn):
    aconn.prepared_max = 5
    aconn.prepare_threshold = 0
    for i in range(10):
        await aconn.execute("select 'a'")
        await aconn.execute(f"select {i}")

    assert len(aconn._prepared._names) == 5
    for j in [9, 8, 7, 6, "'a'"]:
        name = aconn._prepared._names[f"select {j}".encode(), ()]
        assert name.startswith(b"_pg3_")

    stmts = await get_prepared_statements(aconn)
    stmts.sort(key=lambda rec: rec.prepare_time)
    got = [stmt.statement for stmt in stmts]
    assert got == [f"select {i}" for i in ["'a'", 6, 7, 8, 9]]


@pytest.mark.skipif("psycopg._cmodule._psycopg", reason="Python-only debug conn")
async def test_deallocate_or_close(aconn, caplog):
    aconn.pgconn = PGconnDebug(aconn.pgconn)
    caplog.set_level(logging.INFO, logger="psycopg.debug")

    await aconn.set_autocommit(True)
    aconn.prepare_threshold = 0
    aconn.prepared_max = 1

    await aconn.execute("select 1::bigint")
    await aconn.execute("select 1::text")

    msgs = "\n".join(rec.message for rec in caplog.records)
    if psycopg.pq.__build_version__ >= 170000:
        assert "PGconn.send_close_prepared" in msgs
        assert "DEALLOCATE" not in msgs
    else:
        assert "PGconn.send_close_prepared" not in msgs
        assert "DEALLOCATE" in msgs


def test_prepared_max_none(conn):
    conn.prepared_max = 42
    assert conn.prepared_max == 42
    assert conn._prepared.prepared_max == 42

    conn.prepared_max = None
    assert conn._prepared.prepared_max == sys.maxsize
    assert conn.prepared_max is None

    conn.prepared_max = 0
    assert conn._prepared.prepared_max == 0
    assert conn.prepared_max == 0

    conn.prepared_max = 24
    assert conn.prepared_max == 24
    assert conn._prepared.prepared_max == 24


async def test_different_types(aconn):
    aconn.prepare_threshold = 0
    await aconn.execute("select %s", [None])
    await aconn.execute("select %s", [dt.date(2000, 1, 1)])
    await aconn.execute("select %s", [42])
    await aconn.execute("select %s", [41])
    await aconn.execute("select %s", [dt.date(2000, 1, 2)])

    stmts = await get_prepared_statements(aconn)
    stmts.sort(key=lambda rec: rec.prepare_time)
    got = [stmt.parameter_types for stmt in stmts]
    assert got == [["text"], ["date"], ["smallint"]]


async def test_untyped_json(aconn):
    aconn.prepare_threshold = 1
    await aconn.execute("create table testjson(data jsonb)")
    for i in range(2):
        await aconn.execute("insert into testjson (data) values (%s)", ["{}"])

    stmts = await get_prepared_statements(aconn)
    got = [stmt.parameter_types for stmt in stmts]
    assert got == [["jsonb"]]


async def test_change_type_execute(aconn):
    aconn.prepare_threshold = 0
    for i in range(3):
        await aconn.execute("CREATE TYPE prepenum AS ENUM ('foo', 'bar', 'baz')")
        await aconn.execute("CREATE TABLE preptable(id integer, bar prepenum[])")
        await aconn.cursor().execute(
            "INSERT INTO preptable (bar) VALUES (%(enum_col)s::prepenum[])",
            {"enum_col": ["foo"]},
        )
        await aconn.rollback()


async def test_change_type_executemany(aconn):
    for i in range(3):
        await aconn.execute("CREATE TYPE prepenum AS ENUM ('foo', 'bar', 'baz')")
        await aconn.execute("CREATE TABLE preptable(id integer, bar prepenum[])")
        await aconn.cursor().executemany(
            "INSERT INTO preptable (bar) VALUES (%(enum_col)s::prepenum[])",
            [{"enum_col": ["foo"]}, {"enum_col": ["foo", "bar"]}],
        )
        await aconn.rollback()


@pytest.mark.crdb("skip", reason="can't re-create a type")
async def test_change_type(aconn):
    aconn.prepare_threshold = 0
    await aconn.execute("CREATE TYPE prepenum AS ENUM ('foo', 'bar', 'baz')")
    await aconn.execute("CREATE TABLE preptable(id integer, bar prepenum[])")
    await aconn.cursor().execute(
        "INSERT INTO preptable (bar) VALUES (%(enum_col)s::prepenum[])",
        {"enum_col": ["foo"]},
    )
    await aconn.execute("DROP TABLE preptable")
    await aconn.execute("DROP TYPE prepenum")
    await aconn.execute("CREATE TYPE prepenum AS ENUM ('foo', 'bar', 'baz')")
    await aconn.execute("CREATE TABLE preptable(id integer, bar prepenum[])")
    await aconn.cursor().execute(
        "INSERT INTO preptable (bar) VALUES (%(enum_col)s::prepenum[])",
        {"enum_col": ["foo"]},
    )

    stmts = await get_prepared_statements(aconn)
    assert len(stmts) == 3


async def test_change_type_savepoint(aconn):
    aconn.prepare_threshold = 0
    async with aconn.transaction():
        for i in range(3):
            with pytest.raises(ZeroDivisionError):
                async with aconn.transaction():
                    await aconn.execute(
                        "CREATE TYPE prepenum AS ENUM ('foo', 'bar', 'baz')"
                    )
                    await aconn.execute(
                        "CREATE TABLE preptable(id integer, bar prepenum[])"
                    )
                    await aconn.cursor().execute(
                        "INSERT INTO preptable (bar) VALUES (%(enum_col)s::prepenum[])",
                        {"enum_col": ["foo"]},
                    )
                    raise ZeroDivisionError()


async def get_prepared_statements(aconn):
    cur = aconn.cursor(row_factory=namedtuple_row)
    # CRDB has 'PREPARE name AS' in the statement.
    await cur.execute(
        """
select name,
    regexp_replace(statement, 'prepare _pg3_\\d+ as ', '', 'i') as statement,
    prepare_time,
    parameter_types
from pg_prepared_statements
where name != ''
        """,
        prepare=False,
    )
    return await cur.fetchall()



import pytest

import psycopg
from psycopg.conninfo import conninfo_attempts_async, conninfo_to_dict

pytestmark = pytest.mark.anyio


@pytest.mark.parametrize(
    "conninfo, want, env",
    [
        ("", [""], None),
        ("service=foo", ["service=foo"], None),
        ("host='' user=bar", ["host='' user=bar"], None),
        (
            "host=127.0.0.1 user=bar port=''",
            ["host=127.0.0.1 user=bar port='' hostaddr=127.0.0.1"],
            None,
        ),
        (
            "host=127.0.0.1 user=bar",
            ["host=127.0.0.1 user=bar hostaddr=127.0.0.1"],
            None,
        ),
        (
            "host=1.1.1.1,2.2.2.2 user=bar",
            [
                "host=1.1.1.1 user=bar hostaddr=1.1.1.1",
                "host=2.2.2.2 user=bar hostaddr=2.2.2.2",
            ],
            None,
        ),
        (
            "host=1.1.1.1,2.2.2.2 port=5432",
            [
                "host=1.1.1.1 port=5432 hostaddr=1.1.1.1",
                "host=2.2.2.2 port=5432 hostaddr=2.2.2.2",
            ],
            None,
        ),
        (
            "host=1.1.1.1,2.2.2.2 port=5432,",
            [
                "host=1.1.1.1 port=5432 hostaddr=1.1.1.1",
                "host=2.2.2.2 port='' hostaddr=2.2.2.2",
            ],
            None,
        ),
        (
            "host=1.1.1.1,2.2.2.2 target_session_attrs=prefer-standby",
            [
                "host=1.1.1.1 hostaddr=1.1.1.1 target_session_attrs=standby",
                "host=2.2.2.2 hostaddr=2.2.2.2 target_session_attrs=standby",
                "host=1.1.1.1 hostaddr=1.1.1.1",
                "host=2.2.2.2 hostaddr=2.2.2.2",
            ],
            None,
        ),
        (
            "port=5432",
            [
                "host=1.1.1.1 port=5432 hostaddr=1.1.1.1",
                "host=2.2.2.2 port=5432 hostaddr=2.2.2.2",
            ],
            {"PGHOST": "1.1.1.1,2.2.2.2"},
        ),
        (
            "host=foo.com port=5432",
            ["host=foo.com port=5432"],
            {"PGHOSTADDR": "1.2.3.4"},
        ),
    ],
)
async def test_conninfo_attempts_no_resolve(
    setpgenv, conninfo, want, env, fail_resolve
):
    setpgenv(env)
    params = conninfo_to_dict(conninfo)
    attempts = await conninfo_attempts_async(params)
    want = list(map(conninfo_to_dict, want))
    assert want == attempts


@pytest.mark.parametrize(
    "conninfo, want, env",
    [
        (
            "host=foo.com,qux.com",
            ["host=foo.com hostaddr=1.1.1.1", "host=qux.com hostaddr=2.2.2.2"],
            None,
        ),
        (
            "host=foo.com,qux.com port=5433",
            [
                "host=foo.com hostaddr=1.1.1.1 port=5433",
                "host=qux.com hostaddr=2.2.2.2 port=5433",
            ],
            None,
        ),
        (
            "host=foo.com,qux.com port=5432,5433",
            [
                "host=foo.com hostaddr=1.1.1.1 port=5432",
                "host=qux.com hostaddr=2.2.2.2 port=5433",
            ],
            None,
        ),
        (
            "host=foo.com,foo.com port=5432,",
            [
                "host=foo.com hostaddr=1.1.1.1 port=5432",
                "host=foo.com hostaddr=1.1.1.1 port=''",
            ],
            None,
        ),
        (
            "host=foo.com,nosuchhost.com",
            ["host=foo.com hostaddr=1.1.1.1"],
            None,
        ),
        (
            "host=foo.com, port=5432,5433",
            ["host=foo.com hostaddr=1.1.1.1 port=5432", "host='' port=5433"],
            None,
        ),
        (
            "host=nosuchhost.com,foo.com",
            ["host=foo.com hostaddr=1.1.1.1"],
            None,
        ),
        (
            "host=foo.com,qux.com",
            ["host=foo.com hostaddr=1.1.1.1", "host=qux.com hostaddr=2.2.2.2"],
            {},
        ),
        (
            "host=dup.com",
            ["host=dup.com hostaddr=3.3.3.3", "host=dup.com hostaddr=3.3.3.4"],
            None,
        ),
        (
            "host=dup.com target_session_attrs=prefer-standby",
            [
                "host=dup.com hostaddr=3.3.3.3 target_session_attrs=standby",
                "host=dup.com hostaddr=3.3.3.4 target_session_attrs=standby",
                "host=dup.com hostaddr=3.3.3.3",
                "host=dup.com hostaddr=3.3.3.4",
            ],
            None,
        ),
    ],
)
async def test_conninfo_attempts(conninfo, want, env, fake_resolve):
    params = conninfo_to_dict(conninfo)
    attempts = await conninfo_attempts_async(params)
    want = list(map(conninfo_to_dict, want))
    assert want == attempts


@pytest.mark.parametrize(
    "conninfo, env",
    [
        ("host=bad1.com,bad2.com", None),
        ("host=foo.com port=1,2", None),
        ("host=1.1.1.1,2.2.2.2 port=5432,5433,5434", None),
        ("host=1.1.1.1,2.2.2.2", {"PGPORT": "1,2,3"}),
    ],
)
async def test_conninfo_attempts_bad(setpgenv, conninfo, env, fake_resolve):
    setpgenv(env)
    params = conninfo_to_dict(conninfo)
    with pytest.raises(psycopg.Error):
        await conninfo_attempts_async(params)


async def test_conninfo_random_multi_host():
    hosts = [f"host{n:02d}" for n in range(50)]
    args = {"host": ",".join(hosts), "hostaddr": ",".join(["127.0.0.1"] * len(hosts))}
    ahosts = [att["host"] for att in await conninfo_attempts_async(args)]
    assert ahosts == hosts

    args["load_balance_hosts"] = "disable"
    ahosts = [att["host"] for att in await conninfo_attempts_async(args)]
    assert ahosts == hosts

    args["load_balance_hosts"] = "random"
    ahosts = [att["host"] for att in await conninfo_attempts_async(args)]
    assert ahosts != hosts
    ahosts.sort()
    assert ahosts == hosts


async def test_conninfo_random_multi_ips(fake_resolve):
    args = {"host": "alot.com"}
    hostaddrs = [str(att["hostaddr"]) for att in await conninfo_attempts_async(args)]
    assert len(hostaddrs) == 20
    assert hostaddrs == sorted(hostaddrs)

    args["load_balance_hosts"] = "disable"
    hostaddrs = [str(att["hostaddr"]) for att in await conninfo_attempts_async(args)]
    assert hostaddrs == sorted(hostaddrs)

    args["load_balance_hosts"] = "random"
    hostaddrs = [str(att["hostaddr"]) for att in await conninfo_attempts_async(args)]
    assert hostaddrs != sorted(hostaddrs)



"""
Psycopg AsyncCursor object.
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

from types import TracebackType
from typing import TYPE_CHECKING, Any, overload
from contextlib import asynccontextmanager
from collections.abc import AsyncIterator, Iterable

from . import errors as e
from . import pq
from .abc import Params, Query, QueryNoTemplate
from .copy import AsyncCopy, AsyncWriter
from .rows import AsyncRowFactory, Row, RowMaker
from ._compat import Self, Template
from ._cursor_base import BaseCursor
from ._pipeline_async import AsyncPipeline

if TYPE_CHECKING:
    from .connection_async import AsyncConnection

ACTIVE = pq.TransactionStatus.ACTIVE


class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
    __module__ = "psycopg"
    __slots__ = ()

    @overload
    def __init__(self, connection: AsyncConnection[Row]): ...

    @overload
    def __init__(
        self, connection: AsyncConnection[Any], *, row_factory: AsyncRowFactory[Row]
    ): ...

    def __init__(
        self,
        connection: AsyncConnection[Any],
        *,
        row_factory: AsyncRowFactory[Row] | None = None,
    ):
        super().__init__(connection)
        self._row_factory = row_factory or connection.row_factory

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close()

    async def close(self) -> None:
        """
        Close the current cursor and free associated resources.
        """
        self._close()

    @property
    def row_factory(self) -> AsyncRowFactory[Row]:
        """Writable attribute to control how result rows are formed."""
        return self._row_factory

    @row_factory.setter
    def row_factory(self, row_factory: AsyncRowFactory[Row]) -> None:
        self._row_factory = row_factory
        if self.pgresult:
            self._make_row = row_factory(self)

    def _make_row_maker(self) -> RowMaker[Row]:
        return self._row_factory(self)

    @overload
    async def execute(
        self,
        query: QueryNoTemplate,
        params: Params | None = None,
        *,
        prepare: bool | None = None,
        binary: bool | None = None,
    ) -> Self: ...

    @overload
    async def execute(
        self,
        query: Template,
        *,
        prepare: bool | None = None,
        binary: bool | None = None,
    ) -> Self: ...

    async def execute(
        self,
        query: Query,
        params: Params | None = None,
        *,
        prepare: bool | None = None,
        binary: bool | None = None,
    ) -> Self:
        """
        Execute a query or command to the database.
        """
        try:
            async with self._conn.lock:
                await self._conn.wait(
                    self._execute_gen(query, params, prepare=prepare, binary=binary)
                )
        except e._NO_TRACEBACK as ex:
            raise ex.with_traceback(None)
        return self

    async def executemany(
        self, query: Query, params_seq: Iterable[Params], *, returning: bool = False
    ) -> None:
        """
        Execute the same command with a sequence of input data.
        """
        try:
            async with self._conn.lock:
                if AsyncPipeline.is_supported():
                    # If there is already a pipeline, ride it, in order to avoid
                    # sending unnecessary Sync.
                    if self._conn._pipeline:
                        await self._conn.wait(
                            self._executemany_gen_pipeline(query, params_seq, returning)
                        )
                    # Otherwise, make a new one
                    else:
                        async with self._conn._pipeline_nolock():
                            await self._conn.wait(
                                self._executemany_gen_pipeline(
                                    query, params_seq, returning
                                )
                            )
                else:
                    await self._conn.wait(
                        self._executemany_gen_no_pipeline(query, params_seq, returning)
                    )
        except e._NO_TRACEBACK as ex:
            raise ex.with_traceback(None)

    async def stream(
        self,
        query: Query,
        params: Params | None = None,
        *,
        binary: bool | None = None,
        size: int = 1,
    ) -> AsyncIterator[Row]:
        """
        Iterate row-by-row on a result from the database.

        :param size: if greater than 1, results will be retrieved by chunks of
            this size from the server (but still yielded row-by-row); this is only
            available from version 17 of the libpq.
        """
        if self._pgconn.pipeline_status:
            raise e.ProgrammingError("stream() cannot be used in pipeline mode")

        async with self._conn.lock:
            try:
                await self._conn.wait(
                    self._stream_send_gen(query, params, binary=binary, size=size)
                )
                first = True
                while res := await self._conn.wait(self._stream_fetchone_gen(first)):
                    for pos in range(res.ntuples):
                        yield self._tx.load_row(pos, self._make_row)
                    first = False
            except e._NO_TRACEBACK as ex:
                raise ex.with_traceback(None)
            finally:
                if self._pgconn.transaction_status == ACTIVE:
                    # Try to cancel the query, then consume the results
                    # already received.
                    await self._conn._try_cancel()
                    try:
                        while await self._conn.wait(
                            self._stream_fetchone_gen(first=False)
                        ):
                            pass
                    except Exception:
                        pass

                    # Try to get out of ACTIVE state. Just do a single attempt, which
                    # should work to recover from an error or query cancelled.
                    try:
                        await self._conn.wait(self._stream_fetchone_gen(first=False))
                    except Exception:
                        pass

    async def results(self) -> AsyncIterator[Self]:
        """
        Iterate across multiple record sets received by the cursor.

        Multiple record sets are received after using `executemany()` with
        `!returning=True` or using `execute()` with more than one query in the
        command.
        """
        if self.pgresult:
            while True:
                yield self

                if not self.nextset():
                    break

    async def set_result(self, index: int) -> Self:
        """
        Move to a specific result set.

        :arg index: index of the result to go to
        :type index: `!int`

        More than one result will be available after executing calling
        `executemany()` or `execute()` with more than one query.

        `!index` is 0-based and supports negative values, counting from the end,
        the same way you can index items in a list.

        The function returns self, so that the result may be followed by a
        fetch operation. See `results()` for details.
        """
        if not -len(self._results) <= index < len(self._results):
            raise IndexError(
                f"index {index} out of range: {len(self._results)} result(s) available"
            )
        if index < 0:
            index = len(self._results) + index

        self._select_current_result(index)
        return self

    async def fetchone(self) -> Row | None:
        """
        Return the next record from the current result set.

        Return `!None` the result set is finished.

        :rtype: Row | None, with Row defined by `row_factory`
        """
        await self._fetch_pipeline()
        res = self._check_result_for_fetch()
        if self._pos < res.ntuples:
            record = self._tx.load_row(self._pos, self._make_row)
            self._pos += 1
            return record
        return None

    async def fetchmany(self, size: int = 0) -> list[Row]:
        """
        Return the next `!size` records from the current result set.

        `!size` default to `!self.arraysize` if not specified.

        :rtype: Sequence[Row], with Row defined by `row_factory`
        """
        await self._fetch_pipeline()
        res = self._check_result_for_fetch()

        if not size:
            size = self.arraysize
        records = self._tx.load_rows(
            self._pos, min(self._pos + size, res.ntuples), self._make_row
        )
        self._pos += len(records)
        return records

    async def fetchall(self) -> list[Row]:
        """
        Return all the remaining records from the current result set.

        :rtype: Sequence[Row], with Row defined by `row_factory`
        """
        await self._fetch_pipeline()
        res = self._check_result_for_fetch()
        records = self._tx.load_rows(self._pos, res.ntuples, self._make_row)
        self._pos = res.ntuples
        return records

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> Row:
        await self._fetch_pipeline()
        res = self._check_result_for_fetch()
        if self._pos < res.ntuples:
            record = self._tx.load_row(self._pos, self._make_row)
            self._pos += 1
            return record
        raise StopAsyncIteration("no more records to return")

    async def scroll(self, value: int, mode: str = "relative") -> None:
        """
        Move the cursor in the result set to a new position according to mode.

        If `!mode` is ``'relative'`` (default), `!value` is taken as offset to
        the current position in the result set; if set to ``'absolute'``,
        `!value` states an absolute target position.

        Raise `!IndexError` in case a scroll operation would leave the result
        set. In this case the position will not change.
        """
        await self._fetch_pipeline()
        self._scroll(value, mode)

    @asynccontextmanager
    async def copy(
        self,
        statement: Query,
        params: Params | None = None,
        *,
        writer: AsyncWriter | None = None,
    ) -> AsyncIterator[AsyncCopy]:
        """
        Initiate a :sql:`COPY` operation and return an object to manage it.
        """
        try:
            async with self._conn.lock:
                await self._conn.wait(self._start_copy_gen(statement, params))

                async with AsyncCopy(self, writer=writer) as copy:
                    yield copy
        except e._NO_TRACEBACK as ex:
            raise ex.with_traceback(None)

        # If a fresher result has been set on the cursor by the Copy object,
        # read its properties (especially rowcount).
        self._select_current_result(0)

    async def _fetch_pipeline(self) -> None:
        if (
            self._execmany_returning is not False
            and not self.pgresult
            and self._conn._pipeline
        ):
            async with self._conn.lock:
                await self._conn.wait(self._conn._pipeline._fetch_gen(flush=True))



"""
Helper object to transform values between Python and PostgreSQL

Python implementation of the object. Use the `_transformer module to import
the right implementation (Python or C). The public place where the object
is exported is `psycopg.adapt` (which we may not use to avoid circular
dependencies problems).
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

from typing import TYPE_CHECKING, Any, DefaultDict, TypeAlias
from collections import defaultdict
from collections.abc import Sequence

from . import abc
from . import errors as e
from . import pq
from .abc import AdaptContext, Buffer, LoadFunc, NoneType, PyFormat
from .rows import Row, RowMaker
from ._oids import INVALID_OID, TEXT_OID
from ._encodings import conn_encoding

if TYPE_CHECKING:
    from .abc import DumperKey  # noqa: F401
    from .adapt import AdaptersMap
    from .pq.abc import PGresult
    from ._connection_base import BaseConnection

DumperCache: TypeAlias = "dict[DumperKey, abc.Dumper]"
OidDumperCache: TypeAlias = dict[int, abc.Dumper]
LoaderCache: TypeAlias = dict[int, abc.Loader]

TEXT = pq.Format.TEXT
PY_TEXT = PyFormat.TEXT


class Transformer(AdaptContext):
    """
    An object that can adapt efficiently between Python and PostgreSQL.

    The life cycle of the object is the query, so it is assumed that attributes
    such as the server version or the connection encoding will not change. The
    object have its state so adapting several values of the same type can be
    optimised.

    """

    __module__ = "psycopg.adapt"

    __slots__ = """
        types formats
        _conn _adapters _pgresult _dumpers _loaders _encoding _none_oid
        _oid_dumpers _oid_types _row_dumpers _row_loaders
        """.split()

    types: tuple[int, ...] | None
    formats: list[pq.Format] | None

    _adapters: AdaptersMap
    _pgresult: PGresult | None
    _none_oid: int

    def __init__(self, context: AdaptContext | None = None):
        self._pgresult = self.types = self.formats = None

        # WARNING: don't store context, or you'll create a loop with the Cursor
        if context:
            self._adapters = context.adapters
            self._conn = context.connection
        else:
            from . import postgres

            self._adapters = postgres.adapters
            self._conn = None

        # mapping fmt, class -> Dumper instance
        self._dumpers: DefaultDict[PyFormat, DumperCache]
        self._dumpers = defaultdict(dict)

        # mapping fmt, oid -> Dumper instance
        # Not often used, so create it only if needed.
        self._oid_dumpers: tuple[OidDumperCache, OidDumperCache] | None
        self._oid_dumpers = None

        # mapping fmt, oid -> Loader instance
        self._loaders: tuple[LoaderCache, LoaderCache] = ({}, {})

        self._row_dumpers: list[abc.Dumper] | None = None

        # sequence of load functions from value to python
        # the length of the result columns
        self._row_loaders: list[LoadFunc] = []

        # mapping oid -> type sql representation
        self._oid_types: dict[int, bytes] = {}

        self._encoding = ""

    @classmethod
    def from_context(cls, context: AdaptContext | None) -> Transformer:
        """
        Return a Transformer from an AdaptContext.

        If the context is a Transformer instance, just return it.
        """
        if isinstance(context, Transformer):
            return context
        else:
            return cls(context)

    @property
    def connection(self) -> BaseConnection[Any] | None:
        return self._conn

    @property
    def encoding(self) -> str:
        if not self._encoding:
            self._encoding = conn_encoding(self.connection)
        return self._encoding

    @property
    def adapters(self) -> AdaptersMap:
        return self._adapters

    @property
    def pgresult(self) -> PGresult | None:
        return self._pgresult

    def set_pgresult(
        self,
        result: PGresult | None,
        *,
        set_loaders: bool = True,
        format: pq.Format | None = None,
    ) -> None:
        self._pgresult = result

        if not result:
            self._nfields = self._ntuples = 0
            if set_loaders:
                self._row_loaders = []
            return

        self._ntuples = result.ntuples
        nf = self._nfields = result.nfields

        if not set_loaders:
            return

        if not nf:
            self._row_loaders = []
            return

        fmt: pq.Format
        fmt = result.fformat(0) if format is None else format  # type: ignore
        self._row_loaders = [
            self.get_loader(result.ftype(i), fmt).load for i in range(nf)
        ]

    def set_dumper_types(self, types: Sequence[int], format: pq.Format) -> None:
        self._row_dumpers = [self.get_dumper_by_oid(oid, format) for oid in types]
        self.types = tuple(types)
        self.formats = [format] * len(types)

    def set_loader_types(self, types: Sequence[int], format: pq.Format) -> None:
        self._row_loaders = [self.get_loader(oid, format).load for oid in types]

    def dump_sequence(
        self, params: Sequence[Any], formats: Sequence[PyFormat]
    ) -> Sequence[Buffer | None]:
        nparams = len(params)
        out: list[Buffer | None] = [None] * nparams

        # If we have dumpers, it means set_dumper_types had been called, in
        # which case self.types and self.formats are set to sequences of the
        # right size.
        if self._row_dumpers:
            if len(self._row_dumpers) != nparams:
                raise e.DataError(
                    f"expected {len(self._row_dumpers)} values in row, got {nparams}"
                )
            for i in range(nparams):
                if (param := params[i]) is not None:
                    out[i] = self._row_dumpers[i].dump(param)
            return out

        types = [self._get_none_oid()] * nparams
        pqformats = [TEXT] * nparams

        for i in range(nparams):
            if (param := params[i]) is None:
                continue
            dumper = self.get_dumper(param, formats[i])
            out[i] = dumper.dump(param)
            types[i] = dumper.oid
            pqformats[i] = dumper.format

        self.types = tuple(types)
        self.formats = pqformats

        return out

    def as_literal(self, obj: Any) -> bytes:
        dumper = self.get_dumper(obj, PY_TEXT)
        rv = dumper.quote(obj)
        # If the result is quoted, and the oid not unknown or text,
        # add an explicit type cast.
        # Check the last char because the first one might be 'E'.
        oid = dumper.oid
        if oid and rv and rv[-1] == b"'"[0] and oid != TEXT_OID:
            try:
                type_sql = self._oid_types[oid]
            except KeyError:
                if ti := self.adapters.types.get(oid):
                    if oid < 8192:
                        # builtin: prefer "timestamptz" to "timestamp with time zone"
                        type_sql = ti.name.encode(self.encoding)
                    else:
                        type_sql = ti.regtype.encode(self.encoding)
                    if oid == ti.array_oid:
                        type_sql += b"[]"
                else:
                    type_sql = b""
                self._oid_types[oid] = type_sql

            if type_sql:
                rv = b"%s::%s" % (rv, type_sql)

        if not isinstance(rv, bytes):
            rv = bytes(rv)
        return rv

    def get_dumper(self, obj: Any, format: PyFormat) -> abc.Dumper:
        """
        Return a Dumper instance to dump `!obj`.
        """
        # Normally, the type of the object dictates how to dump it
        key = type(obj)

        # Reuse an existing Dumper class for objects of the same type
        cache = self._dumpers[format]
        try:
            dumper = cache[key]
        except KeyError:
            # If it's the first time we see this type, look for a dumper
            # configured for it.
            try:
                dcls = self.adapters.get_dumper(key, format)
            except e.ProgrammingError as ex:
                raise ex from None
            else:
                cache[key] = dumper = dcls(key, self)

        # Check if the dumper requires an upgrade to handle this specific value
        if (key1 := dumper.get_key(obj, format)) is key:
            return dumper

        # If it does, ask the dumper to create its own upgraded version
        try:
            return cache[key1]
        except KeyError:
            dumper = cache[key1] = dumper.upgrade(obj, format)
            return dumper

    def _get_none_oid(self) -> int:
        try:
            return self._none_oid
        except AttributeError:
            pass

        try:
            rv = self._none_oid = self._adapters.get_dumper(NoneType, PY_TEXT).oid
        except KeyError:
            raise e.InterfaceError("None dumper not found")

        return rv

    def get_dumper_by_oid(self, oid: int, format: pq.Format) -> abc.Dumper:
        """
        Return a Dumper to dump an object to the type with given oid.
        """
        if not self._oid_dumpers:
            self._oid_dumpers = ({}, {})

        # Reuse an existing Dumper class for objects of the same type
        cache = self._oid_dumpers[format]
        try:
            return cache[oid]
        except KeyError:
            # If it's the first time we see this type, look for a dumper
            # configured for it.
            dcls = self.adapters.get_dumper_by_oid(oid, format)
            cache[oid] = dumper = dcls(NoneType, self)

        return dumper

    def load_rows(self, row0: int, row1: int, make_row: RowMaker[Row]) -> list[Row]:
        if not (res := self._pgresult):
            raise e.InterfaceError("result not set")

        if not (0 <= row0 <= self._ntuples and 0 <= row1 <= self._ntuples):
            raise e.InterfaceError(
                f"rows must be included between 0 and {self._ntuples}"
            )

        records = []
        for row in range(row0, row1):
            record: list[Any] = [None] * self._nfields
            for col in range(self._nfields):
                if (val := res.get_value(row, col)) is not None:
                    record[col] = self._row_loaders[col](val)
            records.append(make_row(record))

        return records

    def load_row(self, row: int, make_row: RowMaker[Row]) -> Row:
        if not (res := self._pgresult):
            raise e.InterfaceError("result not set")

        if not 0 <= row <= self._ntuples:
            raise e.InterfaceError(
                f"row must be included between 0 and {self._ntuples}"
            )

        record: list[Any] = [None] * self._nfields
        for col in range(self._nfields):
            if (val := res.get_value(row, col)) is not None:
                record[col] = self._row_loaders[col](val)

        return make_row(record)

    def load_sequence(self, record: Sequence[Buffer | None]) -> tuple[Any, ...]:
        if len(self._row_loaders) != len(record):
            raise e.ProgrammingError(
                f"cannot load sequence of {len(record)} items:"
                f" {len(self._row_loaders)} loaders registered"
            )

        return tuple(
            (self._row_loaders[i](val) if val is not None else None)
            for i, val in enumerate(record)
        )

    def get_loader(self, oid: int, format: pq.Format) -> abc.Loader:
        try:
            return self._loaders[format][oid]
        except KeyError:
            pass

        if not (loader_cls := self._adapters.get_loader(oid, format)):
            if not (loader_cls := self._adapters.get_loader(INVALID_OID, format)):
                raise e.InterfaceError("unknown oid loader not found")
        loader = self._loaders[format][oid] = loader_cls(oid, self)
        return loader



"""
Mappings between PostgreSQL and Python encodings.
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

import re
import codecs
import string
from typing import TYPE_CHECKING, Any
from functools import cache

from .errors import NotSupportedError
from .pq._enums import ConnStatus

if TYPE_CHECKING:
    from ._connection_base import BaseConnection

OK = ConnStatus.OK


_py_codecs = {
    "BIG5": "big5",
    "EUC_CN": "gb2312",
    "EUC_JIS_2004": "euc_jis_2004",
    "EUC_JP": "euc_jp",
    "EUC_KR": "euc_kr",
    # "EUC_TW": not available in Python
    "GB18030": "gb18030",
    "GBK": "gbk",
    "ISO_8859_5": "iso8859-5",
    "ISO_8859_6": "iso8859-6",
    "ISO_8859_7": "iso8859-7",
    "ISO_8859_8": "iso8859-8",
    "JOHAB": "johab",
    "KOI8R": "koi8-r",
    "KOI8U": "koi8-u",
    "LATIN1": "iso8859-1",
    "LATIN10": "iso8859-16",
    "LATIN2": "iso8859-2",
    "LATIN3": "iso8859-3",
    "LATIN4": "iso8859-4",
    "LATIN5": "iso8859-9",
    "LATIN6": "iso8859-10",
    "LATIN7": "iso8859-13",
    "LATIN8": "iso8859-14",
    "LATIN9": "iso8859-15",
    # "MULE_INTERNAL": not available in Python
    "SHIFT_JIS_2004": "shift_jis_2004",
    "SJIS": "shift_jis",
    # this actually means no encoding, see PostgreSQL docs
    # it is special-cased by the text loader.
    "SQL_ASCII": "ascii",
    "UHC": "cp949",
    "UTF8": "utf-8",
    "WIN1250": "cp1250",
    "WIN1251": "cp1251",
    "WIN1252": "cp1252",
    "WIN1253": "cp1253",
    "WIN1254": "cp1254",
    "WIN1255": "cp1255",
    "WIN1256": "cp1256",
    "WIN1257": "cp1257",
    "WIN1258": "cp1258",
    "WIN866": "cp866",
    "WIN874": "cp874",
}

py_codecs: dict[bytes, str] = {}
py_codecs.update((k.encode(), v) for k, v in _py_codecs.items())

# Add an alias without underscore, for lenient lookups
py_codecs.update(
    (k.replace("_", "").encode(), v) for k, v in _py_codecs.items() if "_" in k
)

pg_codecs = {v: k.encode() for k, v in _py_codecs.items()}


def conn_encoding(conn: BaseConnection[Any] | None) -> str:
    """
    Return the Python encoding name of a psycopg connection.

    Default to utf8 if the connection has no encoding info.
    """
    return conn.pgconn._encoding if conn else "utf-8"


def conninfo_encoding(conninfo: str) -> str:
    """
    Return the Python encoding name passed in a conninfo string. Default to utf8.

    Because the input is likely to come from the user and not normalised by the
    server, be somewhat lenient (non-case-sensitive lookup, ignore noise chars).
    """
    from .conninfo import conninfo_to_dict

    params = conninfo_to_dict(conninfo)
    if pgenc := params.get("client_encoding"):
        try:
            return pg2pyenc(str(pgenc).encode())
        except NotSupportedError:
            pass

    return "utf-8"


@cache
def py2pgenc(name: str) -> bytes:
    """Convert a Python encoding name to PostgreSQL encoding name.

    Raise LookupError if the Python encoding is unknown.
    """
    return pg_codecs[codecs.lookup(name).name]


@cache
def pg2pyenc(name: bytes) -> str:
    """Convert a PostgreSQL encoding name to Python encoding name.

    Raise NotSupportedError if the PostgreSQL encoding is not supported by
    Python.
    """
    try:
        return py_codecs[name.replace(b"-", b"").replace(b"_", b"").upper()]
    except KeyError:
        sname = name.decode("utf8", "replace")
        raise NotSupportedError(f"codec not available in Python: {sname!r}")


def _as_python_identifier(s: str, prefix: str = "f") -> str:
    """
    Reduce a string to a valid Python identifier.

    Replace all non-valid chars with '_' and prefix the value with `!prefix` if
    the first letter is an '_'.
    """
    if not s.isidentifier():
        if s[0] in "1234567890":
            s = prefix + s
        if not s.isidentifier():
            s = _re_clean.sub("_", s)
    # namedtuple fields cannot start with underscore. So...
    if s[0] == "_":
        s = prefix + s
    return s


_re_clean = re.compile(
    f"[^{string.ascii_lowercase}{string.ascii_uppercase}{string.digits}_]"
)



"""
Support module for test_connection[_async].py
"""

from __future__ import annotations

from typing import Any
from dataclasses import dataclass

import pytest

import psycopg
from psycopg.conninfo import conninfo_to_dict

try:
    from psycopg.conninfo import _DEFAULT_CONNECT_TIMEOUT as DEFAULT_TIMEOUT
except ImportError:
    # Allow tests to import (not necessarily to pass all) if the psycopg module
    # imported is not the one expected (e.g. running psycopg pool tests on the
    # master branch with psycopg 3.1.x imported).
    DEFAULT_TIMEOUT = 130


@pytest.fixture
def testctx(svcconn):
    svcconn.execute("create table if not exists testctx (id int primary key)")
    svcconn.execute("delete from testctx")
    return None


@dataclass
class ParamDef:
    name: str
    guc: str
    values: list[Any]
    non_default: str


param_isolation = ParamDef(
    name="isolation_level",
    guc="isolation",
    values=list(psycopg.IsolationLevel),
    non_default="serializable",
)
param_read_only = ParamDef(
    name="read_only",
    guc="read_only",
    values=[True, False],
    non_default="on",
)
param_deferrable = ParamDef(
    name="deferrable",
    guc="deferrable",
    values=[True, False],
    non_default="on",
)

# Map Python values to Postgres values for the tx_params possible values
tx_values_map = {
    v.name.lower().replace("_", " "): v.value for v in psycopg.IsolationLevel
}
tx_values_map["on"] = True
tx_values_map["off"] = False


tx_params = [
    param_isolation,
    param_read_only,
    pytest.param(param_deferrable, marks=pytest.mark.crdb_skip("deferrable")),
]
tx_params_isolation = [
    pytest.param(
        param_isolation,
        id="isolation_level",
        marks=pytest.mark.crdb("skip", reason="transaction isolation"),
    ),
    pytest.param(
        param_read_only, id="read_only", marks=pytest.mark.crdb_skip("begin_read_only")
    ),
    pytest.param(
        param_deferrable, id="deferrable", marks=pytest.mark.crdb_skip("deferrable")
    ),
]


conninfo_params_timeout = [
    (
        "",
        {"dbname": "mydb", "connect_timeout": None},
        ({"dbname": "mydb"}, DEFAULT_TIMEOUT),
    ),
    (
        "",
        {"dbname": "mydb", "connect_timeout": 1},
        ({"dbname": "mydb", "connect_timeout": 1}, 2),
    ),
    (
        "dbname=postgres",
        {},
        ({"dbname": "postgres"}, DEFAULT_TIMEOUT),
    ),
    (
        "dbname=postgres connect_timeout=2",
        {},
        ({"dbname": "postgres", "connect_timeout": "2"}, 2),
    ),
    (
        "postgresql:///postgres?connect_timeout=2",
        {"connect_timeout": 10},
        ({"dbname": "postgres", "connect_timeout": 10}, 10),
    ),
]


def drop_default_args_from_conninfo(conninfo):
    if isinstance(conninfo, str):
        params = conninfo_to_dict(conninfo)
    else:
        params = conninfo.copy()

    def removeif(key, value):
        if params.get(key) == value:
            params.pop(key)

    removeif("connect_timeout", str(DEFAULT_TIMEOUT))

    return params



"""
The Column object in Cursor.description
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

from typing import TYPE_CHECKING, Any
from operator import attrgetter
from collections.abc import Sequence

if TYPE_CHECKING:
    from ._cursor_base import BaseCursor


class Column(Sequence[Any]):
    __module__ = "psycopg"

    def __init__(self, cursor: BaseCursor[Any, Any], index: int):
        res = cursor.pgresult
        assert res

        if fname := res.fname(index):
            self._name = fname.decode(cursor._encoding)
        else:
            # COPY_OUT results have columns but no name
            self._name = f"column_{index + 1}"

        self._ftype = res.ftype(index)
        self._type = cursor.adapters.types.get(self._ftype)
        self._fmod = res.fmod(index)
        self._fsize = res.fsize(index)

    _attrs = tuple(attrgetter(attr) for attr in """
            name type_code display_size internal_size precision scale null_ok
            """.split())

    def __repr__(self) -> str:
        return (
            f"<Column {self.name!r},"
            f" type: {self.type_display} (oid: {self.type_code})>"
        )

    def __len__(self) -> int:
        return 7

    @property
    def type_display(self) -> str:
        """A pretty representation of the column type.

        It is composed by the type name, followed by eventual modifiers and
        brackets to signify arrays, e.g. :sql:`text`, :sql:`varchar(42)`,
        :sql:`date[]`.
        """
        if not self._type:
            return str(self.type_code)

        return self._type.get_type_display(oid=self.type_code, fmod=self._fmod)

    def __getitem__(self, index: Any) -> Any:
        if isinstance(index, slice):
            return tuple(getter(self) for getter in self._attrs[index])
        else:
            return self._attrs[index](self)

    @property
    def name(self) -> str:
        """The name of the column."""
        return self._name

    @property
    def type_code(self) -> int:
        """The numeric OID of the column."""
        return self._ftype

    @property
    def display_size(self) -> int | None:
        """The field size, for string types such as :sql:`varchar(n)`."""
        return self._type.get_display_size(self._fmod) if self._type else None

    @property
    def internal_size(self) -> int | None:
        """The internal field size for fixed-size types, None otherwise."""
        fsize = self._fsize
        return fsize if fsize >= 0 else None

    @property
    def precision(self) -> int | None:
        """The number of digits for fixed precision types."""
        return self._type.get_precision(self._fmod) if self._type else None

    @property
    def scale(self) -> int | None:
        """The number of digits after the decimal point if available."""
        return self._type.get_scale(self._fmod) if self._type else None

    @property
    def null_ok(self) -> bool | None:
        """Always `!None`"""
        return None



import socket
import asyncio

import pytest


@pytest.fixture
def fake_resolve(monkeypatch):
    """
    Fixture to return known name from name resolution.
    """
    fake_hosts = {
        "localhost": ["127.0.0.1"],
        "foo.com": ["1.1.1.1"],
        "qux.com": ["2.2.2.2"],
        "dup.com": ["3.3.3.3", "3.3.3.4"],
        "alot.com": [f"4.4.4.{n}" for n in range(10, 30)],
    }

    def family(host):
        return socket.AF_INET6 if ":" in host else socket.AF_INET

    def fake_getaddrinfo(host, port, *args, **kwargs):
        assert isinstance(port, int) or (isinstance(port, str) and port.isdigit())
        try:
            addrs = fake_hosts[host]
        except KeyError:
            raise OSError(f"unknown test host: {host}")
        else:
            return [
                (family(addr), socket.SOCK_STREAM, 6, "", (addr, port))
                for addr in addrs
            ]

    _patch_gai(monkeypatch, fake_getaddrinfo)


@pytest.fixture
def fail_resolve(monkeypatch):
    """
    Fixture to fail any name resolution.
    """

    def fail_getaddrinfo(host, port, **kwargs):
        pytest.fail(f"shouldn't try to resolve {host}")

    _patch_gai(monkeypatch, fail_getaddrinfo)


def _patch_gai(monkeypatch, f):
    monkeypatch.setattr(socket, "getaddrinfo", f)
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        pass
    else:

        async def af(*args, **kwargs):
            return f(*args, **kwargs)

        monkeypatch.setattr(loop, "getaddrinfo", af)



# WARNING: this file is auto-generated by 'async_to_sync.py'
# from the original file 'test_copy_async.py'
# DO NOT CHANGE! Change the original file instead.
import string
from random import choice, randrange
from typing import Any  # noqa: ignore

import pytest

from psycopg import errors as e
from psycopg import pq, sql
from psycopg.adapt import PyFormat
from psycopg.types.numeric import Int4

from ..utils import eur
from .._test_copy import sample_binary  # noqa
from .._test_copy import ensure_table, sample_records
from .._test_copy import sample_tabledef as sample_tabledef_pg
from .._test_copy import sample_text

# CRDB int/serial are int8
sample_tabledef = sample_tabledef_pg.replace("int", "int4").replace("serial", "int4")

pytestmark = [pytest.mark.crdb]


@pytest.mark.parametrize(
    "format, buffer",
    [(pq.Format.TEXT, "sample_text"), (pq.Format.BINARY, "sample_binary")],
)
def test_copy_in_buffers(conn, format, buffer):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with cur.copy(f"copy copy_in from stdin {copyopt(format)}") as copy:
        copy.write(globals()[buffer])

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


def test_copy_in_buffers_pg_error(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(e.UniqueViolation):
        with cur.copy("copy copy_in from stdin") as copy:
            copy.write(sample_text)
            copy.write(sample_text)
    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


def test_copy_in_str(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with cur.copy("copy copy_in from stdin") as copy:
        copy.write(sample_text.decode())

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


@pytest.mark.xfail(reason="bad sqlstate - CRDB #81559")
def test_copy_in_error(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(e.QueryCanceled):
        with cur.copy("copy copy_in from stdin with binary") as copy:
            copy.write(sample_text.decode())

    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_empty(conn, format):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with cur.copy(f"copy copy_in from stdin {copyopt(format)}"):
        pass

    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS
    assert cur.rowcount == 0


@pytest.mark.slow
def test_copy_big_size_record(conn):
    cur = conn.cursor()
    ensure_table(cur, "id serial primary key, data text")
    data = "".join((chr(randrange(1, 256)) for i in range(10 * 1024 * 1024)))
    with cur.copy("copy copy_in (data) from stdin") as copy:
        copy.write_row([data])

    cur.execute("select data from copy_in limit 1")
    assert cur.fetchone()[0] == data


@pytest.mark.slow
def test_copy_big_size_block(conn):
    cur = conn.cursor()
    ensure_table(cur, "id serial primary key, data text")
    data = "".join((choice(string.ascii_letters) for i in range(10 * 1024 * 1024)))
    copy_data = data + "\n"
    with cur.copy("copy copy_in (data) from stdin") as copy:
        copy.write(copy_data)

    cur.execute("select data from copy_in limit 1")
    assert cur.fetchone()[0] == data


def test_copy_in_buffers_with_pg_error(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(e.UniqueViolation):
        with cur.copy("copy copy_in from stdin") as copy:
            copy.write(sample_text)
            copy.write(sample_text)

    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_records(conn, format):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)

    with cur.copy(f"copy copy_in from stdin {copyopt(format)}") as copy:
        row: tuple[Any, ...]
        for row in sample_records:
            if format == pq.Format.BINARY:
                row = tuple((Int4(i) if isinstance(i, int) else i for i in row))
            copy.write_row(row)

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_records_set_types(conn, format):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)

    with cur.copy(f"copy copy_in from stdin {copyopt(format)}") as copy:
        copy.set_types(["int4", "int4", "text"])
        for row in sample_records:
            copy.write_row(row)

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_records_binary(conn, format):
    cur = conn.cursor()
    ensure_table(cur, "col1 serial primary key, col2 int4, data text")

    with cur.copy(f"copy copy_in (col2, data) from stdin {copyopt(format)}") as copy:
        for row in sample_records:
            copy.write_row((None, row[2]))

    cur.execute("select col2, data from copy_in order by 2")
    data = cur.fetchall()
    assert data == [(None, "hello"), (None, "world")]


@pytest.mark.crdb_skip("copy canceled")
def test_copy_in_buffers_with_py_error(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(e.QueryCanceled) as exc:
        with cur.copy("copy copy_in from stdin") as copy:
            copy.write(sample_text)
            raise Exception("nuttengoggenio")

    assert "nuttengoggenio" in str(exc.value)
    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


def test_copy_in_allchars(conn):
    cur = conn.cursor()
    ensure_table(cur, "col1 int primary key, col2 int, data text")

    with cur.copy("copy copy_in from stdin") as copy:
        for i in range(1, 256):
            copy.write_row((i, None, chr(i)))
        copy.write_row((ord(eur), None, eur))

    cur.execute("""
select col1 = ascii(data), col2 is null, length(data), count(*)
from copy_in group by 1, 2, 3
""")
    data = cur.fetchall()
    assert data == [(True, True, 1, 256)]


@pytest.mark.slow
@pytest.mark.parametrize(
    "fmt, set_types",
    [(pq.Format.TEXT, True), (pq.Format.TEXT, False), (pq.Format.BINARY, True)],
)
@pytest.mark.crdb_skip("copy array")
def test_copy_from_leaks(conn_cls, dsn, faker, fmt, set_types, gc):
    faker.format = PyFormat.from_pq(fmt)
    faker.choose_schema(ncols=20)
    faker.make_records(20)

    def work():
        with conn_cls.connect(dsn) as conn:
            with conn.cursor(binary=fmt) as cur:
                cur.execute(faker.drop_stmt)
                cur.execute(faker.create_stmt)

                stmt = sql.SQL("copy {} ({}) from stdin {}").format(
                    faker.table_name,
                    sql.SQL(", ").join(faker.fields_names),
                    sql.SQL("with binary" if fmt else ""),
                )
                with cur.copy(stmt) as copy:
                    if set_types:
                        copy.set_types(faker.types_names)
                    for row in faker.records:
                        copy.write_row(row)

                cur.execute(faker.select_stmt)
                recs = cur.fetchall()

                for got, want in zip(recs, faker.records):
                    faker.assert_record(got, want)

    gc.collect()
    n = []
    for i in range(3):
        work()
        gc.collect()
        n.append(gc.count())

    assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"


def copyopt(format):
    return "with binary" if format == pq.Format.BINARY else ""



# WARNING: this file is auto-generated by 'async_to_sync.py'
# from the original file 'test_notify_async.py'
# DO NOT CHANGE! Change the original file instead.
from __future__ import annotations

from time import time

import pytest

from psycopg import Notify

from .utils import skip_free_threaded
from .acompat import Event, gather, sleep, spawn

pytestmark = pytest.mark.crdb_skip("notify")


def test_notify_handlers(conn):
    nots1 = []
    nots2 = []

    def cb1(n):
        nots1.append(n)

    conn.add_notify_handler(cb1)
    conn.add_notify_handler(lambda n: nots2.append(n))

    conn.set_autocommit(True)
    conn.execute("listen foo")
    conn.execute("notify foo, 'n1'")

    assert len(nots1) == 1
    n = nots1[0]
    assert n.channel == "foo"
    assert n.payload == "n1"
    assert n.pid == conn.pgconn.backend_pid

    assert len(nots2) == 1
    assert nots2[0] == nots1[0]

    conn.remove_notify_handler(cb1)
    conn.execute("notify foo, 'n2'")

    assert len(nots1) == 1
    assert len(nots2) == 2
    n = nots2[1]
    assert isinstance(n, Notify)
    assert n.channel == "foo"
    assert n.payload == "n2"
    assert n.pid == conn.pgconn.backend_pid
    assert hash(n)

    with pytest.raises(ValueError):
        conn.remove_notify_handler(cb1)


@pytest.mark.slow
@pytest.mark.timing
def test_notify(conn_cls, conn, dsn):
    npid = None

    def notifier():
        with conn_cls.connect(dsn, autocommit=True) as nconn:
            nonlocal npid
            npid = nconn.pgconn.backend_pid

            sleep(0.25)
            nconn.execute("notify foo, '1'")
            sleep(0.25)
            nconn.execute("notify foo, '2'")

    def receiver():
        conn.set_autocommit(True)
        cur = conn.cursor()
        cur.execute("listen foo")
        gen = conn.notifies()
        for n in gen:
            ns.append((n, time()))
            if len(ns) >= 2:
                gen.close()

    ns: list[tuple[Notify, float]] = []
    t0 = time()
    workers = [spawn(notifier), spawn(receiver)]
    gather(*workers)
    assert len(ns) == 2

    n, t1 = ns[0]
    assert n.pid == npid
    assert n.channel == "foo"
    assert n.payload == "1"
    assert t1 - t0 == pytest.approx(0.25, abs=0.05)

    n, t1 = ns[1]
    assert n.pid == npid
    assert n.channel == "foo"
    assert n.payload == "2"
    assert t1 - t0 == pytest.approx(0.5, abs=0.05)


@pytest.mark.slow
@pytest.mark.timing
def test_no_notify_timeout(conn):
    conn.set_autocommit(True)
    t0 = time()
    for n in conn.notifies(timeout=0.5):
        assert False
    dt = time() - t0
    assert 0.5 <= dt < 0.75


@pytest.mark.slow
@pytest.mark.timing
def test_notify_timeout(conn_cls, conn, dsn):
    conn.set_autocommit(True)
    conn.execute("listen foo")

    def notifier():
        with conn_cls.connect(dsn, autocommit=True) as nconn:
            sleep(0.25)
            nconn.execute("notify foo, '1'")

    worker = spawn(notifier)
    try:
        times = [time()]
        for n in conn.notifies(timeout=0.5):
            times.append(time())
        times.append(time())
    finally:
        gather(worker)

    assert len(times) == 3
    assert times[1] - times[0] == pytest.approx(0.25, 0.1)
    assert times[2] - times[1] == pytest.approx(0.25, 0.1)


@pytest.mark.slow
@pytest.mark.timing
def test_notify_timeout_0(conn_cls, conn, dsn):
    conn.set_autocommit(True)
    conn.execute("listen foo")

    ns = list(conn.notifies(timeout=0))
    assert not ns

    with conn_cls.connect(dsn, autocommit=True) as nconn:
        nconn.execute("notify foo, '1'")
        sleep(0.1)

    ns = list(conn.notifies(timeout=0))
    assert len(ns) == 1


@pytest.mark.slow
@pytest.mark.timing
def test_stop_after(conn_cls, conn, dsn):
    conn.set_autocommit(True)
    conn.execute("listen foo")

    def notifier():
        with conn_cls.connect(dsn, autocommit=True) as nconn:
            nconn.execute("notify foo, '1'")
            sleep(0.1)
            nconn.execute("notify foo, '2'")
            sleep(0.1)
            nconn.execute("notify foo, '3'")

    worker = spawn(notifier)
    try:
        ns = list(conn.notifies(timeout=1.0, stop_after=2))
        assert len(ns) == 2
        assert ns[0].payload == "1"
        assert ns[1].payload == "2"
    finally:
        gather(worker)

    ns = list(conn.notifies(timeout=0.0))
    assert len(ns) == 1
    assert ns[0].payload == "3"


@pytest.mark.timing
def test_stop_after_batch(conn_cls, conn, dsn):
    conn.set_autocommit(True)
    conn.execute("listen foo")

    def notifier():
        with conn_cls.connect(dsn, autocommit=True) as nconn:
            with nconn.transaction():
                nconn.execute("notify foo, '1'")
                nconn.execute("notify foo, '2'")

    worker = spawn(notifier)
    try:
        ns = list(conn.notifies(timeout=1.0, stop_after=1))
        assert len(ns) == 2
        assert ns[0].payload == "1"
        assert ns[1].payload == "2"
    finally:
        gather(worker)


@pytest.mark.slow
@pytest.mark.timing
def test_notifies_blocking(conn):

    def listener():
        for _ in conn.notifies(timeout=1):
            pass

    worker = spawn(listener)
    try:
        # Make sure the listener is listening
        if not conn.lock.locked():
            sleep(0.01)

        t0 = time()
        conn.execute("select 1")
        dt = time() - t0
    finally:
        gather(worker)

    assert dt > 0.5


@pytest.mark.slow
@skip_free_threaded("warnings are context-local in the free-threaded build >= 3.14")
def test_generator_and_handler(conn, conn_cls, dsn, recwarn):
    # NOTE: we don't support generator+handlers anymore. So, if in the future
    # this behaviour will change, we will not consider it a regression. However
    # we will want to keep the warning check.

    recwarn.clear()

    conn.set_autocommit(True)
    conn.execute("listen foo")

    n1 = None
    n2 = None

    def set_n2(n):
        nonlocal n2
        n2 = n

    conn.add_notify_handler(set_n2)

    def listener():
        nonlocal n1
        for n1 in conn.notifies(timeout=1, stop_after=1):
            pass

    worker = spawn(listener)
    try:
        # Make sure the listener is listening
        if not conn.lock.locked():
            sleep(0.01)

        with conn_cls.connect(dsn, autocommit=True) as nconn:
            nconn.execute("notify foo, '1'")
    finally:
        gather(worker)

    assert n1
    assert n2

    msg = str(recwarn.pop(RuntimeWarning).message)
    assert "notifies()" in msg


@pytest.mark.parametrize("query_between", [True, False])
def test_first_notify_not_lost(conn, conn_cls, dsn, query_between):
    conn.set_autocommit(True)
    conn.execute("listen foo")

    with conn_cls.connect(dsn, autocommit=True) as conn2:
        conn2.execute("notify foo, 'hi'")

    if query_between:
        conn.execute("select 1")

    n = None
    for n in conn.notifies(timeout=1, stop_after=1):
        pass
    assert n


@pytest.mark.slow
@pytest.mark.timing
@pytest.mark.parametrize("sleep_on", ["server", "client"])
@pytest.mark.parametrize("listen_by", ["callback", "generator"])
def test_notify_query_notify(conn_cls, dsn, sleep_on, listen_by):
    e = Event()
    notifies: list[int] = []
    workers = []

    def notifier():
        with conn_cls.connect(dsn, autocommit=True) as conn:
            sleep(0.1)
            for i in range(3):
                conn.execute("select pg_notify('counter', %s)", (str(i),))
                sleep(0.2)

    def nap(conn):
        if sleep_on == "server":
            conn.execute("select pg_sleep(0.2)")
        else:
            assert sleep_on == "client"
            sleep(0.2)

    if listen_by == "callback":

        def listener():
            with conn_cls.connect(dsn, autocommit=True) as conn:
                conn.add_notify_handler(lambda n: notifies.append(int(n.payload)))

                conn.execute("listen counter")
                e.set()

                nap(conn)
                conn.execute("")
                nap(conn)
                conn.execute("")
                nap(conn)
                conn.execute("")

    else:

        def listener():
            with conn_cls.connect(dsn, autocommit=True) as conn:
                conn.execute("listen counter")
                e.set()
                for n in conn.notifies(timeout=0.2):
                    notifies.append(int(n.payload))

                nap(conn)

                for n in conn.notifies(timeout=0.2):
                    notifies.append(int(n.payload))

    workers.append(spawn(listener))
    e.wait()
    workers.append(spawn(notifier))
    gather(*workers)

    assert notifies == list(range(3))
